{-# LANGUAGE BangPatterns, TypeSynonymInstances, TupleSections, FlexibleInstances #-}
module FVXL where
import Control.Monad
import Data.Maybe
import Data.Char
import Data.Set (Set)
import qualified Data.Set as S
import Data.Vec (Vec3D(..), Vec3I(..), Vec2I(..), dot, cross, pack, unpack, normalize)
import qualified Data.Vec as V
import Data.Fixed (mod')
import Data.Array.Unboxed
import Data.Char
import Text.Printf
import Text.JSON
import qualified Data.MessagePack as MP
import Data.Ix
import Data.List
import Debug.Trace
import Data.Bits
import Data.Word
import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as BSL
import Data.Binary.Put

-- TODO: move Vec3xx to TraceVoxel.
-- drop dependency on Data.Vec
shapeOf :: IArray a e => a Vec3I e -> Vec3I
shapeOf arr = V.Vec3I 1 1 1 + (high-low)
	where (low,high) = bounds arr

instance Ix Vec2I where
	range (i0,i1) = map tupleTov2i $ range (v2iToTuple i0, v2iToTuple i1)
	index (i0,i1) i = index (v2iToTuple i0, v2iToTuple i1) (v2iToTuple i)
	inRange (i0,i1) i = Data.Ix.inRange (v2iToTuple i0, v2iToTuple i1) (v2iToTuple i)

v2iToTuple (V.Vec2I i j) = (i,j)
tupleTov2i (i,j) = V.Vec2I i j

instance Ix Vec3I where
	range (i0,i1) = map tupleTov3i $ range (v3iToTuple i0, v3iToTuple i1)
	index (i0,i1) i = index (v3iToTuple i0, v3iToTuple i1) (v3iToTuple i)
	inRange (i0,i1) i = Data.Ix.inRange (v3iToTuple i0, v3iToTuple i1) (v3iToTuple i)

v3iToTuple (V.Vec3I i j k) = (i,j,k)
tupleTov3i (i,j,k) = V.Vec3I i j k



putSimpleFVXL :: UArray Vec3I Word8 -> Put
putSimpleFVXL arr = do
	mapM_ (putWord32be . fromIntegral) [nx,ny,nz]
	putWord8 $ fromIntegral $ ord 'o'
	mapM_ putWord8 $ elems arr
	where
		V.Vec3I nx ny nz = shapeOf arr

putFVXL :: FVXL -> Put
putFVXL fvxl@(FVXL fields values) 
	|isFVXLValid fvxl = do
		putByteString $ BS.pack $ map (fromIntegral . ord) "fvxl2013"
		putWord8 1 -- 1 type of channels
		putField $ (Field 'o' 1 Continuous U8)
	|otherwise = error "unrealizable FVXL data"

isFVXLValid :: FVXL -> Bool
isFVXLValid (FVXL fields values) = True

putField :: Field -> Put
putField (Field fc fnum fm ft) = do
	putWord8 $ fromIntegral $ ord fc
	putWord8 $ fromIntegral fnum
	putWord8 $ cvMode fm .|. cvType ft
	where
		cvMode Discrete = 0x80
		cvMode Continuous = 0x00
		cvType U8 = 0x11
		cvType U16 = 0x12
		cvType U32 = 0x14
		cvType F16 = 0x22
		cvType F32 = 0x34
		cvType F64 = 0x28

data FVXL = FVXL [Field] [Either (UArray Vec3I Word32) (UArray Vec3I Double)]

-- | 'Field' is combination of, 'FieldMode' and 'FieldType'.
-- Any combination of 'FieldMode' and 'FieldType' is possible.
-- Discrete float 'Field's are a little bit unusual, but they occur in some cases.
data Field = Field !Char !Int !FieldMode !FieldType

-- | 'Discrete' values doesn't have valid meaning when interpolated.
-- Continuous values should be interpolated when possible.
-- discrete:0x80 continuos:0x00
data FieldMode = Discrete | Continuous

data FieldType 
	= U8  -- ^ unsigned 8 bit integer : 0x11
	| U16 -- ^ unsigned 16 bit integer : 0x12
	| U32 -- ^ unsigned 32 bit integer : 0x14
	| F16 -- ^ IEEE half float : 0x22
	| F32 -- ^ IEEE single : 0x24
	| F64 -- ^ IEEE double : 0x28


dumpSimpleFVXL :: FilePath -> UArray Vec3I Word8 -> IO ()
dumpSimpleFVXL path arr = BSL.writeFile path $ runPut $ putSimpleFVXL arr


