{-# LANGUAGE BangPatterns, TypeSynonymInstances, TupleSections #-}
-- | Spherical Geometry
module Spherical where
import Data.Maybe
import Data.Vec (Vec3D(..), Vec3I(..), Vec2I(..), dot, cross, pack, unpack, normalize)
import qualified Data.Vec as V
import Data.Fixed (mod')
import Data.List
import Debug.Trace


data Range a = Range !a !a deriving(Show,Eq)

inRange :: Ord a => Range a -> a -> Bool
inRange (Range x0 x1) x = x0<=x && x<=x1


data SphericalAngle = SphericalAngle !(Range Double) !(Range Double) deriving(Show)

saContains :: SphericalAngle -> (Double,Double) -> Bool
saContains (SphericalAngle tr pr) (t,p) = inRange tr t && inRange pr p

-- | Return 'Triangle's whose union truly contains given 'SphericalAngle'.
saToTris :: SphericalAngle -> [Triangle]
saToTris (SphericalAngle tr pr)
	|t0<1e-6 && t1>pi-1e-6 = error "implement: this should return two tris"
	|t0<1e-6 = [triangle pNorth p01 p11]
	|t1>pi-1e-6 = [triangle pSouth p10 p00]
	|otherwise = [triangle p00 p01 p11, triangle p11 p10 p00]
	where
		arcP0 = lonArc p0
		arcP1 = lonArc p1
		circT0 = largerC t0 True
		circT1 = largerC t1 False

		p00 = fromJust $ intersectArcCircle arcP0 circT0
		p01 = fromJust $ intersectArcCircle arcP0 circT1
		p10 = fromJust $ intersectArcCircle arcP1 circT0
		p11 = fromJust $ intersectArcCircle arcP1 circT1
		pNorth = V.Vec3D 0 0 1
		pSouth = V.Vec3D 0 0 (-1)

		Range p0 p1 = pr
		Range t0 t1 = tr
		phiCenter = (p0+p1)/2

		tangentialC t = extend $ arcFromNearest (spToR3 t $ phiCenter) (spToR3 (0.5*pi) (phiCenter+0.5*pi))
		vertexC t = extend $ arcFromNearest (spToR3 t p0) (spToR3 t p1)
		largerC t sideZero
			|sideZero == (t<0.5*pi) = vertexC t
			|otherwise = tangentialC t
		lonArc p = Arc (spToR3 (0.5*pi) (p+0.5*pi)) (V.Vec3D 0 0 1) (-1)

saCenter :: SphericalAngle -> Vec3D
saCenter (SphericalAngle (Range t0 t1) (Range p0 p1)) = spToR3 ((t0+t1)/2) ((p0+p1)/2)

saMaxRadius :: SphericalAngle -> Double
saMaxRadius (SphericalAngle (Range t0 t1) (Range p0 p1)) = (t1-t0)/2 + (p1-p0)/2

spToR3 :: Double -> Double -> Vec3D
spToR3 t p = V.Vec3D (sin t*cos p) (sin t*sin p) (cos t)


-- | Arc of great circle. always shorter than or equal to pi, so cos(angle) is enough.
data Arc = Arc !Vec3D !Vec3D !Double deriving(Show) -- ^ n, p0, cos(angle)

data Circle = Circle !Vec3D deriving(Show) -- ^ n


-- | cross for packed vector
crossP :: Vec3D -> Vec3D -> Vec3D
crossP p q = pack $ unpack p `cross` unpack q

-- | Create 'Arc' connecting given 'Vec3D's in the shortest path.
-- The 'Arc' will always have <pi angle. if p and q is polar, the result is undefined.
arcFromNearest :: Vec3D -> Vec3D -> Arc
arcFromNearest p q = Arc n p c
	where
		c = p `dot` q
		n = V.normalize ns
		ns = p `crossP` q

-- | = toAngle arc v <= a
{-# INLINE onArc #-}
onArc :: Arc -> Vec3D -> Bool
onArc arc@(Arc n p ca) v = s>=0 && c>=ca
	where 
		c = p `dot` v
		s = n `dot` (p `crossP` v)

intersectArc :: Arc -> Arc -> Maybe Vec3D
intersectArc arc0@(Arc n0 p0 _) arc1@(Arc n1 p1 _)
	|isValid pole = Just pole
	|isValid apole = Just apole
	|otherwise = Nothing
	where
		isValid p = onArc arc0 p && onArc arc1 p
		pole = normalize $ n0 `crossP` n1 -- pack $ normalize $ unpack n0 `cross` unpack n1
		apole = V.map negate pole

intersectArcCircle :: Arc -> Circle -> Maybe Vec3D
intersectArcCircle arc0@(Arc n0 p0 _) (Circle n1)
	|isValid pole = Just pole
	|isValid apole = Just apole
	|otherwise = Nothing
	where
		isValid p = onArc arc0 p
		pole = pack $ normalize $ unpack n0 `cross` unpack n1
		apole = V.map negate pole


arcSide :: Arc -> Vec3D -> Bool
arcSide (Arc n _ _) v = v `dot` n >=0

extend :: Arc -> Circle
extend (Arc n _ _) = Circle n

data Triangle = Triangle !Vec3D !Arc !Arc !Arc deriving(Show)

triangle :: Vec3D -> Vec3D -> Vec3D -> Triangle
triangle p q r = Triangle p (arcFromNearest p q) (arcFromNearest q r) (arcFromNearest r p)

triContains :: Triangle -> Vec3D -> Bool
triContains (Triangle p a0 a1 a2) v = all (flip arcSide v) [a0,a1,a2]

intersectTriangle :: Triangle -> Triangle -> Bool
intersectTriangle t0@(Triangle p a0 a1 a2) t1@(Triangle q b0 b1 b2) =
	any isJust [intersectArc a b | a<-[a0,a1,a2], b<-[b0,b1,b2]] ||
	triContains t0 q ||
	triContains t1 p
