{-# LANGUAGE TemplateHaskell #-}
import Test.Framework.TH
import Test.Framework (defaultMain, testGroup)
import Test.Framework.Providers.HUnit
import Test.HUnit
import qualified TraceVoxel as TV

import Control.Monad
import Data.Maybe
import qualified Data.Vec as V
import qualified Data.Set as S
import qualified Data.Map as M
import Data.Ix
import Data.Array
import qualified Data.Array.Unboxed as AU

main = $(defaultMainGenerator)

case_shaft_center = mapM_ centerIsContained $ elems $ TV.generateShaftTable (V.Vec3I 1 1 1, V.Vec3I 1 1 1) 5
	where
		centerIsContained shaft = assertBool "non-empty" $ shaft AU.! (V.Vec3I 0 0 0)

-- | All cells in a shaft must be traversable by "directional fill".
case_shaft_connected = mapM_ shaftIsConnected $ elems $ TV.generateShaftTable (V.Vec3I 1 1 1, V.Vec3I 1 1 1) 3
	where
		shaftIsConnected shaft = traverse shaft S.empty (S.singleton (V.Vec3I 0 0 0)) @?= collectAll shaft

		collectAll shaft = S.fromList $ filter (shaft AU.!) $ range $ AU.bounds shaft

		traverse shaft visited frontier
			|S.null frontier = visited
			|otherwise = traverse shaft visited' (frontier' `S.difference` visited')
			where
				visited' = frontier `S.union` visited
				frontier' = S.fromList $ filter (canMoveTo shaft) $ concatMap expandFrom $ S.toList frontier

		canMoveTo shaft i = inRange (AU.bounds shaft) i && (shaft AU.! i)

		expandFrom i@(V.Vec3I x y z) = catMaybes [xp,xm,yp,ym,zp,zm]
			where
				xp = if x>=0 then Just (i+V.Vec3I 1 0 0) else Nothing
				xm = if x<=0 then Just (i-V.Vec3I 1 0 0) else Nothing
				yp = if y>=0 then Just (i+V.Vec3I 0 1 0) else Nothing
				ym = if y<=0 then Just (i-V.Vec3I 0 1 0) else Nothing
				zp = if z>=0 then Just (i+V.Vec3I 0 0 1) else Nothing
				zm = if z<=0 then Just (i-V.Vec3I 0 0 1) else Nothing


case_principle_direction = mapM_ doesn'tHaveBehind $ elems shaftTable
	where
		doesn'tHaveBehind shaft = mapM_ (\(i,v)->when (isBehind i) $ assertBool "empty" $ not v) $ AU.assocs shaft
			where
				isBehind i = i `V.dot` dir < 0
				dir = TV.shaftDirection shaft

		shaftTable = TV.generateShaftTable (V.Vec3I 1 1 1, V.Vec3I 1 2 3) 3

case_principle_directions = mapM_ shaftIsCovered $ assocs shaftTable
	where
		shaftIsCovered (shaftIndex,shaft) = mapM_ (shaftIsCoveredIn shaftIndex shaft) $ TV.shaftDirections shaft
		shaftIsCoveredIn shaftIndex shaft dir = do
			mapM_ (\(i,v)->when (isBehind i) $ assertBool (show ("empty",dir,shaftIndex,i)) $ not v) $ AU.assocs shaft
			where isBehind i = i `V.dot` dir < 0

		shaftTable = TV.generateShaftTable (V.Vec3I 1 1 1, V.Vec3I 1 1 1) 3



-- | Union of all shafts fills whole space
case_shaft_completeness = do
	assertBool "completeness" $ all inAny $ range bnd
	where
		inAny i = any (AU.!i) $ elems shaftTable
		bnd = AU.bounds $ shaftTable!(V.Vec2I 0 0)
		shaftTable = TV.generateShaftTable (V.Vec3I 1 1 1, V.Vec3I 1 2 3) 3

case_voxel = do
	assertBool "empty voxel" $ TV.isVoxelEmpty emptyCell
	assertBool "half voxel" $ not $ TV.isVoxelEmpty nonEmptyCell
	assertBool "filled voxel"  $ not $ TV.isVoxelEmpty wallCell

case_minimum_distance = do
	-- overlapping
	TV.mindCube (V.Vec3I 2 3 4) (V.Vec3I 2 3 4) @?= 0.0

	-- plane-plane
	TV.mindCube (V.Vec3I 2 3 4) (V.Vec3I 4 3 4) @?= 1.0 -- plus X
	TV.mindCube (V.Vec3I 2 3 4) (V.Vec3I 0 3 4) @?= 1.0 -- minux X
	TV.mindCube (V.Vec3I 2 3 4) (V.Vec3I 3 3 4) @?= 0.0

	-- edge-edge
	TV.mindCube (V.Vec3I 0 0 0) (V.Vec3I 2 2 0) @?= sqrt 2

case_maximum_distance = do
	-- overlapping (diagonal line of 1 cell)
	TV.maxdCube (V.Vec3I 2 3 4) (V.Vec3I 2 3 4) @?= sqrt 3

	-- 6-neighbor (diagonal line of 2 cells)
	TV.maxdCube (V.Vec3I 2 3 4) (V.Vec3I 2 3 5) @?= sqrt (1^2+1^2+2^2)


-- | Check low-level collector
case_collect3_raw = do
	regions_lb @?= S.singleton (V.Vec3I 2 2 2)
	regions_ub @?= S.singleton (V.Vec3I 2 2 2, 0)
	where
		regions_lb = TV.collectCCForLB (V.Vec3I 2 2 2) shaft cube3
		regions_ub = TV.collectCCForUB (V.Vec3I 2 2 2) shaft cube3

		-- Shaft is defined in relative coordinates. Let's use half space x>=0 here.
		shaft = AU.listArray bnd $ map isInside $ range bnd 
		isInside (V.Vec3I x _ _) = x>=0
		bnd = (V.Vec3I (-3) (-3) (-3), V.Vec3I 3 3 3)

case_collect3 = do
	TV.shaftDirection shaft @?= V.Vec3I 1 0 0

	cells_lb @?= S.fromList [
		V.Vec3I 2 2 1, V.Vec3I 2 2 3, -- Y neighbor
		V.Vec3I 2 1 2, V.Vec3I 2 3 2, -- Z neighbor
		V.Vec3I 3 2 2] -- X neighbor
	cells_ub @?= S.fromList [V.Vec3I 2 2 2]
	where
		cells_lb = TV.collectTargetsLB (V.Vec3I 2 2 2) shaft cube3
		cells_ub = TV.collectTargetsUB (V.Vec3I 2 2 2) shaft cube3

		-- Shaft is defined in relative coordinates. Let's use half space x>=0 here.
		shaft = AU.listArray bnd $ map isInside $ range bnd 
		isInside (V.Vec3I x _ _) = x>=0
		bnd = (V.Vec3I (-3) (-3) (-3), V.Vec3I 3 3 3)

case_collect4 = do
	TV.shaftDirection shaft @?= V.Vec3I 1 0 0

	cells_lb @?= S.fromList [
		V.Vec3I 2 2 1, V.Vec3I 2 2 3, -- Y neighbor
		V.Vec3I 2 1 2, V.Vec3I 2 3 2, -- Z neighbor
		V.Vec3I 3 2 2] -- X neighbor
	cells_ub @?= S.fromList [V.Vec3I 3 2 2] -- half cell!
	where
		cells_lb = TV.collectTargetsLB (V.Vec3I 2 2 2) shaft cube4
		cells_ub = TV.collectTargetsUB (V.Vec3I 2 2 2) shaft cube4

		-- Shaft is defined in relative coordinates. Let's use half space x>=0 here.
		shaft = AU.listArray bnd $ map isInside $ range bnd 
		isInside (V.Vec3I x _ _) = x>=0
		bnd = (V.Vec3I (-4) (-3) (-3), V.Vec3I 4 3 3)

case_collect4h = do
	TV.shaftDirection shaft @?= V.Vec3I 1 0 0

	cells_lb @?= S.fromList [
		V.Vec3I 2 2 1, V.Vec3I 2 2 3, -- Y neighbor
		V.Vec3I 2 1 2, V.Vec3I 2 3 2, -- Z neighbor
		V.Vec3I 3 2 2] -- X neighbor
	cells_ub @?= S.fromList [V.Vec3I 2 2 2] -- It can't move into non-connected half-filled cell!
	where
		cells_lb = TV.collectTargetsLB (V.Vec3I 2 2 2) shaft cube4h
		cells_ub = TV.collectTargetsUB (V.Vec3I 2 2 2) shaft cube4h

		-- Shaft is defined in relative coordinates. Let's use half space x>=0 here.
		shaft = AU.listArray bnd $ map isInside $ range bnd 
		isInside (V.Vec3I x _ _) = x>=0
		bnd = (V.Vec3I (-4) (-3) (-3), V.Vec3I 4 3 3)

case_collect4h2 = do
	TV.shaftDirection shaft @?= V.Vec3I 1 0 0

	cells_lb @?= S.fromList [
		V.Vec3I 2 2 1, V.Vec3I 2 2 3, -- Y neighbor
		V.Vec3I 2 1 2, V.Vec3I 2 3 2, -- Z neighbor
		V.Vec3I 3 2 2] -- X neighbor
	cells_ub @?= S.fromList [V.Vec3I 3 2 2] -- ray moves into 2 region cell and stop there
	where
		cells_lb = TV.collectTargetsLB (V.Vec3I 2 2 2) shaft cube4h2
		cells_ub = TV.collectTargetsUB (V.Vec3I 2 2 2) shaft cube4h2

		-- Shaft is defined in relative coordinates. Let's use half space x>=0 here.
		shaft = AU.listArray bnd $ map isInside $ range bnd 
		isInside (V.Vec3I x _ _) = x>=0
		bnd = (V.Vec3I (-4) (-3) (-3), V.Vec3I 4 3 3)

case_cube_333 = do
	-- check size
	bounds ddf @?= (V.Vec3I 1 1 1, V.Vec3I 3 3 3)
	-- check existence of DDSphere
	mapM_ (\(i,v)->when (i/=V.Vec3I 2 2 2) $ assertBool "cellType" $ isNothing v) $ assocs ddf
	assertBool "cellType2" $ isJust $ ddf!(V.Vec3I 2 2 2)

	-- check size of (the only) DDSphere
	bounds dds @?= (V.Vec2I 0 0, V.Vec2I (3-1) (2*3-1)) -- ntheta,nphi
	-- check length of some directions
	mapM_ (\(lb,_)->assertBool "Lower Bound" $ lb<1e-3) $ elems dds -- since there's always neighboring wall, lower bound should be close to 0
	mapM_ (\(_,ub)->assertBool "Upper Bound" $ ub>sqrt 3.0 - 1e-3) $ elems dds -- it's always larger than 1-cell straight across
	where
		Just (TV.DDSphere dds) = ddf!(V.Vec3I 2 2 2)
		TV.DDF ddf = TV.generateDDF 3 cube3

case_cube_433 = do
	mapM_ (\(lb,_)->assertBool "Lower Bound" $ lb<1e-3) $ elems dds -- occuluded by half-cell at (3,2,2)
	mapM_ (\(_,ub)->assertBool "Upper Bound" $ ub>sqrt 3.0 -1e-3) $ elems dds -- all cell should have 1+ UB
	assertBool "Upper Bound(Pass)" $ any (\(_,ub)->ub>sqrt (1+1+2)-1e-3) $ elems dds -- exists two-cell long path
	where
		Just (TV.DDSphere dds) = ddf!(V.Vec3I 2 2 2)
		TV.DDF ddf = TV.generateDDF 3 cube4


-- | Very basic 3^3 cube, where the center cell is empty and others are completely filled.
cube3 :: TV.TVolume
cube3 = TV.TVolume $ listArray bnd $ map cell $ range bnd
	where
		bnd = (V.Vec3I 1 1 1, V.Vec3I 3 3 3)

		cell (V.Vec3I 2 2 2) = emptyCell
		cell _ = wallCell

-- | 4*3*3 cube, where (2,2,2) is empty, (3,2,2) is half-filled, and others are filled
cube4 :: TV.TVolume
cube4 = TV.TVolume $ listArray bnd $ map cell $ range bnd
	where
		bnd = (V.Vec3I 1 1 1, V.Vec3I 4 3 3)

		cell (V.Vec3I 2 2 2) = TV.TVoxel [TV.TVRegion True $ TV.Faces S.empty (S.singleton 0) S.empty S.empty S.empty S.empty]
		cell (V.Vec3I 3 2 2) = TV.TVoxel [TV.TVRegion False $ TV.Faces (S.singleton 0) S.empty S.empty S.empty S.empty S.empty]
		cell _ = wallCell

-- | 4*3*3 where (2,2,2) is empty, (3,2,2) is half-filled, and (2,2,2)-(3,2,2) is not connected.
-- (e.g. there's a wall in X- side of (3,2,2)) Other cells are filled.
-- (*,2,2) will be like |***|   |*  |***|
cube4h :: TV.TVolume
cube4h = TV.TVolume $ listArray bnd $ map cell $ range bnd
	where
		bnd = (V.Vec3I 1 1 1, V.Vec3I 4 3 3)

		cell (V.Vec3I 2 2 2) = TV.TVoxel [TV.TVRegion True $ TV.Faces S.empty S.empty S.empty S.empty S.empty S.empty]
		cell (V.Vec3I 3 2 2) = TV.TVoxel [TV.TVRegion False $ TV.Faces S.empty S.empty S.empty S.empty S.empty S.empty]
		cell _ = wallCell

-- | 4*3*3 where (2,2,2) is empty, (3,2,2) have two-region.
-- (*,2,2) will be like |***|   | * |***|
cube4h2 :: TV.TVolume
cube4h2 = TV.TVolume $ listArray bnd $ map cell $ range bnd
	where
		bnd = (V.Vec3I 1 1 1, V.Vec3I 4 3 3)

		cell (V.Vec3I 2 2 2) = TV.TVoxel [TV.TVRegion True $ TV.Faces S.empty (S.singleton 0) S.empty S.empty S.empty S.empty]
		cell (V.Vec3I 3 2 2) = TV.TVoxel [
			TV.TVRegion False $ TV.Faces (S.singleton 0) S.empty S.empty S.empty S.empty S.empty,
			TV.TVRegion False $ TV.Faces S.empty S.empty S.empty S.empty S.empty S.empty]
		cell _ = wallCell



wallCell = TV.TVoxel []
nonEmptyCell = TV.TVoxel [TV.TVRegion False $ TV.Faces S.empty S.empty S.empty S.empty S.empty S.empty]
emptyCell = TV.TVoxel [TV.TVRegion True $ TV.Faces S.empty S.empty S.empty S.empty S.empty S.empty]
