{-# LANGUAGE BangPatterns, TypeSynonymInstances, TupleSections, FlexibleInstances #-}
module TraceVoxel where
import Control.Monad
import Control.Parallel.Strategies (withStrategy, parListChunk, rdeepseq)
import Control.DeepSeq
import Data.Maybe
import Data.Set (Set)
import qualified Data.Set as S
import Data.Vec (Vec3D(..), Vec3I(..), Vec2I(..), dot, cross, pack, unpack, normalize)
import qualified Data.Vec as V
import Data.Fixed (mod')
import Data.Array.Unboxed
import Data.Char
import Text.Printf
import Text.JSON
import qualified Data.MessagePack as MP
import Data.Ix
import Data.List
import Debug.Trace
import Data.Bits
import Data.Word
import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as BSL
import System.IO.Unsafe
import Data.Binary.Put
import System.Environment
import Spherical
import FVXL

approxContained :: Vec3I -> SphericalAngle -> Maybe Bool
approxContained !d !sa 
	|nd<1e-3 = Just True
	|saContains sa (theta,phi) = Just True
	|nd<2 = Nothing
	|cos_d<cos_l = Just False
	|otherwise = Nothing
	where
		dreal = V.map realToFrac $ unpack d
		nd = V.norm dreal
		dnorm@(V.Vec3D dx dy dz) = pack $ V.map (/nd) dreal
		theta = acos dz
		phi = atan2 dy dx
		margin = atan2 (sqrt 3) nd
		cos_d = dnorm `dot` saCenter sa
		cos_l = cos $ saMaxRadius sa + margin



{-# INLINE isVoxelContained #-}
isVoxelContained :: Vec3I -> SphericalAngle -> Bool
isVoxelContained !d !sa = any id [intersectTriangle t0 t1 | t0<-cu_tris, t1<-sa_tris]
	where
		dr = pack $ V.map realToFrac $ unpack d
		cu_tris = map (triToSTri . mapTri (+dr)) tris_base

		triToSTri (v0,v1,v2)
			|isCCW = triangle nv0 nv1 nv2 
			|otherwise = triangle nv2 nv1 nv0
			where
				isCCW = nv0 `dot` (nv1 `crossP` nv2)>0 -- == nv1 `dot` ((nv1-nv0) `cross` (nv2-nv0))
				nv0 = normalize v0
				nv1 = normalize v1
				nv2 = normalize v2
		mapTri f (v0,v1,v2) = (f v0, f v1, f v2)
		
		sa_tris = saToTris sa

-- | triangles of cube [-1,1]^3
tris_base :: [(Vec3D,Vec3D,Vec3D)]
tris_base = cu0_tris
	where
		cu0_tris = concat [
			quadTwo (V.Vec3D 1 0 0) (V.Vec3D 0 1 0) (V.Vec3D 0 0 1),
			quadTwo (V.Vec3D 0 1 0) (V.Vec3D 0 0 1) (V.Vec3D 1 0 0),
			quadTwo (V.Vec3D 0 0 1) (V.Vec3D 1 0 0) (V.Vec3D 0 1 0)]
		quadTwo a0 a1 a2 = quad a0 a1 a2 ++ quad (V.map negate a0) (V.map negate a2) (V.map negate a1)
		quad a0 a1 a2 = [
			(a0+a1+a2, a0-a1+a2, a0-a1-a2),
			(a0-a1-a2, a0+a1-a2, a0+a1+a2)]

voxelShaft :: Vec3I -> SphericalAngle -> UArray Vec3I Bool
voxelShaft !half_shape !sa = createCacheTable (-half_shape,half_shape) isContained
	where
		isContained d =
			case approxContained d sa of
				Nothing -> isVoxelContained d sa
				Just r -> r


-- | All shafts have principle directions. This is used when calculating LB cells.
-- (cells behind are eliminated)
shaftDirection :: UArray Vec3I Bool -> V.Vec3I
shaftDirection shaft = head $ shaftDirections shaft

-- | Return directions such that intersection of corresponding half spaces truly contain the given shape.
-- All shaft is supposed to have at least one such axis.
shaftDirections :: UArray Vec3I Bool -> [V.Vec3I]
shaftDirections shaft = mapMaybe direction [xp,yp,zp]
	where
		direction ap
			|all (\i->if i `dot` ap<0 then not (shaft!i) else True) (range bSmall) = Just ap
			|all (\i->if i `dot` ap>0 then not (shaft!i) else True) (range bSmall) = Just (-ap)
			|otherwise = Nothing
			where
				bSmall = (V.Vec3I (-1) (-1) (-1), V.Vec3I 1 1 1)

		xp = V.Vec3I 1 0 0
		yp = V.Vec3I 0 1 0
		zp = V.Vec3I 0 0 1

createCacheTable :: (Vec3I,Vec3I) -> (Vec3I -> Bool) -> UArray Vec3I Bool 
createCacheTable bnd f = listArray bnd $ map f $ range bnd



-- | Calculate minimum distance between points in cubes.
--
-- First, two cubes are fused by taking Minkowski difference.
-- Then axis are flipped, and magic happens.
mindCube :: Vec3I -> Vec3I -> Double
mindCube !v0 !v1 = aux $ V.map (realToFrac . abs) (v1-v0) - V.Vec3D 1 1 1
	where aux dv = V.norm $ V.map (max 0) dv

-- | Calculate maximum distance between points in cubes.
maxdCube :: Vec3I -> Vec3I -> Double
maxdCube !v0 !v1 = V.norm $ dv+dd
	where
		dv = V.map realToFrac $ v1-v0
		dd = V.map (signum . (+0.1)) dv


-- | Collect (weakly) connected component of ix for upper bound calculation.
collectCCForUB :: Vec3I -> UArray Vec3I Bool -> TVolume -> Set (Vec3I,Int)
collectCCForUB !ix !shaft !geom@(TVolume arr) = traverseCC S.empty $ S.singleton (ix,0)
	where
		traverseCC :: Set (Vec3I,Int) -> Set (Vec3I,Int) -> Set (Vec3I,Int)
		traverseCC !visited !frontier
			|S.null frontier = visited
			|otherwise = traverseCC (frontier `S.union` visited) (frontier' `S.difference` visited)
			where
				frontier' = S.unions $ map expand $ S.toList frontier

				expand :: (Vec3I,Int) -> Set (Vec3I,Int)
				expand v@(i,r) = vs'
					where
						vs' = S.unions $ map (\(i',rs')->S.map (i',) rs') $ filterFaces (\(i',rs')->isForward i' && isInShaft i') $ zipFaces faceNexts faces
						TVRegion _ faces = geom !!! v

						faceNexts = mapFaces (i+) faceNormals
						isForward i' = norm1 (i'-ix) > norm1 (i-ix)
						isInShaft i' = shaft!(i'-ix)

-- | Collect (weakly) connected component of ix for lower bound calculation.
-- Since all nodes in LB graph is empty, the result is 'Set' 'Vec3I'.
--
-- 'TVoxel' at ix must be empty (have only one 'TVRegion').
collectCCForLB :: Vec3I -> UArray Vec3I Bool -> TVolume -> Set Vec3I
collectCCForLB !ix !shaft !geom@(TVolume arr) = traverseCC S.empty $ S.singleton ix
	where
		traverseCC :: Set Vec3I -> Set Vec3I -> Set Vec3I
		traverseCC !visited !frontier
			|S.null frontier = visited
			|otherwise = traverseCC (frontier `S.union` visited) (frontier' `S.difference` visited)
			where
				frontier' = S.fromList $ concatMap expand $ S.toList frontier

				expand :: Vec3I -> [Vec3I]
				expand i = filter (\i'->isForward i' && isPassable i' && isInShaft i') $ map (i+) directions
					where
						isForward i' = norm1 (i'-ix) > norm1 (i-ix)
						isPassable i' = isVoxelEmpty (arr!i')
						isInShaft i' = shaft!(i'-ix)



collectTargetsUB :: Vec3I -> UArray Vec3I Bool -> TVolume -> Set Vec3I
collectTargetsUB !ix !shaft !geom = S.map fst $ S.filter isSink vs
	where
		isSink :: (Vec3I,Int) -> Bool
		isSink v@(i,_) = S.null $ S.unions $ filterFacesByNormal (\d->isForward d && shaft!(d+i-ix)) faces
			where
				isForward d = norm1 (d+i-ix) > norm1 (i-ix)
				TVRegion _ faces = geom !!! v

		vs = collectCCForUB ix shaft geom

collectTargetsLB :: Vec3I -> UArray Vec3I Bool -> TVolume -> Set Vec3I
collectTargetsLB !ix !shaft !geom@(TVolume arr) = S.filter (\i->isNotEmpty i && isNotBehindShaft i) vs_ns
	where
		isNotEmpty i = not $ isVoxelEmpty $ arr!i
		isNotBehindShaft i = all (\pdir->(i-ix) `dot` pdir >=0) pdirs

		pdirs = shaftDirections shaft
		vs_ns = S.fromList $ concatMap (\v->map (+v) directions) vs
		vs = S.toList $ collectCCForLB ix shaft geom

-- | L1 norm
{-# INLINE norm1 #-}
norm1 :: Vec3I -> Int
norm1 (V.Vec3I x y z) = abs x + abs y + abs z


boundsAt :: Vec3I -> UArray Vec3I Bool -> TVolume -> (Double, Double)
boundsAt ix shaft geom = (lb,ub)
	where
		lb = safeMinimum $ map (mindCube ix) $ S.toList $ collectTargetsLB ix shaft geom
		ub = safeMaximum $ map (maxdCube ix) $ S.toList $ collectTargetsUB ix shaft geom

		safeMinimum xs
			|null xs = unsafePerformIO $ dumpLB >> return 1e6
			|otherwise = minimum xs

		safeMaximum xs
			|null xs = unsafePerformIO $ dumpUB >> return 0
			|otherwise = maximum xs

		dumpLB = do
			printf "no LB at %s\n" $ show ix

		dumpUB = do
			printf "no UB at %s\n" $ show ix

visualizeLB :: Vec3I -> UArray Vec3I Bool -> TVolume -> UArray Vec3I Word8
visualizeLB ix shaft geom@(TVolume arr) = markCells (bounds arr) (collectTargetsLB ix shaft geom) 0 0xff // [(ix,0x80)]

visualizeUB :: Vec3I -> UArray Vec3I Bool -> TVolume -> UArray Vec3I Word8
visualizeUB ix shaft geom@(TVolume arr) =
	markCells (bounds arr) cells 0 0xff // [(ix,0x80)]
	where
		cells = collectTargetsUB ix shaft geom
		cc = S.map fst $ collectCCForUB ix shaft geom

markCells :: (Vec3I,Vec3I) -> Set Vec3I -> Word8 -> Word8 -> UArray Vec3I Word8
markCells bnd cells non_member member = listArray bnd $ map f $ range bnd
	where f i = if S.member i cells then member else non_member

dummyGeom !shape = createCacheTable bnd f
	where
		bnd = (V.Vec3I 0 0 0,hi)
		hi = shape - V.Vec3I 1 1 1
		f (V.Vec3I x y z) = not (2<=x && x<25 && 2<=y && y<48 && 2<=z && z<13)

dRange :: Int -> Range Double -> [Double]
dRange n (Range x0 x1) = [x0+realToFrac i*dx | i<-[0..n-1]]
	where dx = (x1-x0)/realToFrac n


data TVolume = TVolume (Array V.Vec3I TVoxel) deriving(Show)

(TVolume arr) !!! (v,l)  = rs !! l
	where TVoxel rs = arr!v


data TVoxel = TVoxel [TVRegion] deriving(Show)
data TVRegion = TVRegion !Bool !(Faces (S.Set Int)) deriving(Show)

{-# INLINE isVoxelEmpty #-}
isVoxelEmpty :: TVoxel -> Bool
isVoxelEmpty (TVoxel [TVRegion True _])= True
isVoxelEmpty _ = False


data Faces a = Faces !a !a !a !a !a !a deriving(Show)

directions :: [Vec3I]
directions = [
	V.Vec3I (-1) 0 0, V.Vec3I 1 0 0,
	V.Vec3I 0 (-1) 0, V.Vec3I 0 1 0,
	V.Vec3I 0 0 (-1), V.Vec3I 0 0 1]

mapFaces :: (a->b) -> Faces a -> Faces b
mapFaces f (Faces f0 f1 f2 f3 f4 f5) = Faces (f f0) (f f1) (f f2) (f f3) (f f4) (f f5)

mapFacesWithNormal :: ((Vec3I, a) -> b) -> Faces a -> Faces b
mapFacesWithNormal f fs = mapFaces f $ zipFaces faceNormals fs

zipFaces :: Faces a -> Faces b -> Faces (a,b)
zipFaces (Faces a0 a1 a2 a3 a4 a5) (Faces b0 b1 b2 b3 b4 b5)
	= Faces (a0,b0) (a1,b1) (a2,b2) (a3,b3) (a4,b4) (a5,b5)

faceNormals :: Faces Vec3I
faceNormals = Faces 
	(V.Vec3I (-1) 0 0) (V.Vec3I 1 0 0)
	(V.Vec3I 0 (-1) 0) (V.Vec3I 0 1 0)
	(V.Vec3I 0 0 (-1)) (V.Vec3I 0 0 1)

filterFaces :: (a -> Bool) -> Faces a -> [a]
filterFaces f (Faces xm xp ym yp zm zp) = filter f [xm,xp,ym,yp,zm,zp]

filterFacesByNormal :: (Vec3I -> Bool) -> Faces a -> [a]
filterFacesByNormal f faces = map snd $ filterFaces (f . fst) $ zipFaces faceNormals faces



instance JSON TVolume where
	readJSON v = do
		arr <- readJSON v :: Result [[[TVoxel]]]
		return $ TVolume $ listArray (getBounds arr) $ concatMap concat arr
		where
			getBounds a = (
				V.Vec3I 1 1 1,
				V.Vec3I (length a) (length $ head a) (length $ head $ head a))
	showJSON (TVolume arr) = showJSON [[[arr!V.Vec3I ix iy iz | iz<-[1..nz]] | iy<-[1..ny]] | ix<-[1..nx]]
		where
			(V.Vec3I 1 1 1, V.Vec3I nx ny nz) = bounds arr


instance JSON TVoxel where
	readJSON v = liftM TVoxel $ readJSON v
	showJSON (TVoxel rs) = showJSON rs

instance JSON TVRegion where
	readJSON (JSObject dict_r) =
		case (lookup "empty" dict, lookup "connections" dict) of
			(Just e_r, Just conn_r) -> liftM2 TVRegion (readJSON $ e_r) (readJSON conn_r)
		where dict = fromJSObject dict_r
	showJSON (TVRegion empty faces) = JSObject $ toJSObject [
		("empty", showJSON empty),
		("connections", showJSON faces)]

instance JSON a => JSON (Faces a) where
	readJSON v = do
		[f0,f1,f2,f3,f4,f5] <- readJSON v
		return $ Faces f0 f1 f2 f3 f4 f5
	showJSON (Faces f0 f1 f2 f3 f4 f5) = showJSON [f0,f1,f2,f3,f4,f5]


data DDF = DDF (Array Vec3I (Maybe DDSphere)) deriving(Show)


data DDSphere = DDSphere (Array Vec2I (Double,Double)) deriving(Show)

instance NFData DDSphere where
	rnf (DDSphere arr) = rnf arr

instance NFData Vec2I where
	rnf (V.Vec2I x y) = rnf x `seq` rnf y `seq` ()

instance Show Vec2I where
	showsPrec n (V.Vec2I x y)
		|n>=10 = ((printf "(Vec2I %d %d)" x y)++)
		|otherwise = ((printf "Vec2I %d %d" x y)++)


generateDDF :: Int -> TVolume -> DDF
generateDDF n_theta geom@(TVolume arr) = DDF $ listArray bnd $ withStrategy (parListChunk 250 rdeepseq) $ map genDDSphere $ range bnd
	where
		n_phi = 2 * n_theta

		genDDSphere ix@(V.Vec3I x y z)
			|isOK ix = Just $ DDSphere $ amap (\shaft->boundsAt ix shaft geom) shafts
			|otherwise = Nothing

		
		isOK ix = isVoxelEmpty (arr!ix)

		-- TODO: this is a hack! remove this & debug properly
		--isOK ix = isVoxelEmpty (arr!ix) && (all (\d->isVoxelEmpty (arr!(ix+d))) directions)

		shafts = generateShaftTable (bounds arr) n_theta
		bnd = bounds arr

dumpPartialDDF :: Int -> TVolume -> IO ()
dumpPartialDDF n_theta geom@(TVolume arr) = do
	dumpSimpleFVXL "DDF0:0-LB.fvxl" $ listArray bnd $ map (fColor . genPDD) $ range bnd
	where
		n_phi = 2 * n_theta

		fColor x = floor $ 255*x/30

		genPDD ix@(V.Vec3I x y z)
			|isVoxelEmpty (arr!ix) = fst $ boundsAt ix (shafts!ixS) geom
			|otherwise = 0

		ixS = V.Vec2I 1 1

		shafts = generateShaftTable (bounds arr) n_theta
		bnd = bounds arr

{-
    msgpack.pack({
        "samples": index_os.tostring(),
        "grid": {
            "asample": sample_a,
            "lsample": sample_l,
            "valid": int(voxels.sum()),
            "x":nx, "y":ny, "z":nz,
            #"theta":orientation.ntheta,
            #"phi":orientation.nphi,
            #"alpha":orientation.nalpha,
            "lunit":lunit, "aunit":aunit,
            #"offset": map(int, offset),
            "origin": map(float, origin)
        },
        "scene": scene.pack_headerless(),
        "time": time.time(),
        "revision": get_cd_revision()
        }, open(fp, 'wb'))

order: x,y,z,theta,phi,(min,max)


* TraceVoxel.hs
voxel at i: [i-0.5,i+0.5]+?

* scene.py
voxel at i: [i,i+1]+offset 

-}
dumpDDFMP :: V.Vec3D -> FilePath -> FilePath -> Int -> Double -> DDF -> IO ()
dumpDDFMP p0 scene_path path n_theta lunit (DDF arr)
	|n_theta `mod` 2 ==1 = error "n_theta must be even"
	|otherwise = do
		scene <- liftM MP.unpack $ BSL.readFile scene_path
		BSL.writeFile path $ MP.pack $ obj scene
	where
		V.Vec3D p0x p0y p0z = p0 - V.Vec3D lunit lunit lunit

		obj scene = MP.toObject $ MP.Assoc [
			("samples", MP.toObject samples),
			("grid", MP.toObject grid),
			("scene", scene)]

		V.Vec3I nx ny nz = shapeOf arr

		samples = MP.toObject $ concatMap encodeDDSphere $ elems arr

		encodeDDSphere :: Maybe DDSphere -> [Double]
		encodeDDSphere Nothing = packA2 dummyDDS
		encodeDDSphere (Just (DDSphere arr)) = map (*lunit) $ packA2 arr

		dummyDDS = listArray (V.Vec2I 0 0, V.Vec2I (n_theta-1) (2*n_theta-1)) $ repeat (1e6,0)

		packA2 :: Array Vec2I (Double, Double) -> [Double]
		packA2 arr = concatMap (\(dmin,dmax)->[dmin,dmax]) $ elems arr

		grid = MP.toObject $ MP.Assoc [
			("asample",MP.toObject (0.0 :: Double)),
			("lsample",MP.toObject (0.0 :: Double)),
			("aunit",MP.toObject (pi/realToFrac n_theta:: Double)),
			("lunit",MP.toObject lunit),
			("x", MP.toObject nx),
			("y", MP.toObject ny),
			("z", MP.toObject nz),
			("origin", MP.toObject [p0x, p0y, p0z])]


-- | Origin of bnd doesn't matter. just its size.
generateShaftTable :: (Vec3I,Vec3I) -> Int -> Array Vec2I (UArray Vec3I Bool)
generateShaftTable bnd n_theta = listArray bndS $ map genShaft $ range bndS
	where
		n_phi = 2 * n_theta

		genShaft (V.Vec2I ith iph) = voxelShaft (snd bnd-fst bnd+V.Vec3I 2 2 2) sa
			where sa = SphericalAngle
						(Range
							(pi*realToFrac ith/realToFrac n_theta + 1e-3) -- poles are removed to avoid instability
							(pi*realToFrac (ith+1)/realToFrac n_theta - 1e-3))
						(Range
							(2*pi*realToFrac iph/realToFrac n_phi)
							(2*pi*realToFrac (iph+1)/realToFrac n_phi))
		bndS = (V.Vec2I 0 0, V.Vec2I (n_theta-1) (n_phi-1))

dumpAllShaft :: Int -> IO ()
dumpAllShaft n_theta = mapM_ dumpShaftAt $ range (bounds shaftTable)
	where
		dumpShaftAt i@(V.Vec2I it ip) = dumpSimpleFVXL (printf "shaft-%i:%i.fvxl" it ip) $ convert $ shaftTable!i
		convert arr = amap (\x->if x then 0x80 else 0x00) arr

		shaftTable = generateShaftTable (V.Vec3I 0 0 0, V.Vec3I 12 15 18) n_theta


generateDDR :: Int -> TVolume -> Vec2I -> Vec3I -> IO (Maybe (Double,Double))
generateDDR n_theta geom@(TVolume arr) ixS ix
	|isVoxelEmpty (arr!ix) = do
		--mapM_ print $ map toODs $ S.toList $ collectCCForUB ix shaft geom
		--print $ collectTargetsUB ix shaft geom
		--dumpSimpleFVXL "LB.fvxl" $ visualizeLB ix shaft geom
		--dumpSimpleFVXL "UB.fvxl" $ visualizeUB ix shaft geom
		return $ Just $ boundsAt ix shaft geom
	|otherwise = return Nothing
	where
		toODs v@(i,_) = faces -- S.unions $ filterFacesByNormal isForward faces
			where
				isForward d = norm1 (d+i-ix) > norm1 (i-ix)
				TVRegion _ faces = geom !!! v

		shaft = (generateShaftTable (bounds arr) n_theta)!ixS


generateDDS :: Int -> TVolume -> Vec3I -> Maybe DDSphere
generateDDS n_theta geom@(TVolume arr) ix
	|isVoxelEmpty (arr!ix) = Just $ DDSphere $ listArray bndS $ map (\i->boundsAt ix (shaftTable!i) geom) $ range bndS
	|otherwise = Nothing
	where
		bndS = bounds shaftTable
		shaftTable = generateShaftTable (bounds arr) n_theta


-- | Create 'TVolume' from json-type topological voxel format.
loadJSONTV :: FilePath -> IO TVolume
loadJSONTV path = do
	Ok tv <- liftM decode $ readFile path
	return tv


-- | Replace all outer cells with filled region and fix connections accordingly.
fillBoundary :: TVolume -> TVolume
fillBoundary (TVolume arr) = TVolume arr'
	where
		arr' = listArray (bounds arr) $ map f $ assocs arr
		f (i,tv)
			|isBoundary i = TVoxel [] -- filled region
			|otherwise =  appXm $ appXp $ appYm $ appYp $ appZm $ appZp tv
			where
				appXm = applyIf (isXMinusIn i) (disconnectToward (V.Vec3I (-1) 0 0))
				appXp = applyIf (isXPlusIn i) (disconnectToward (V.Vec3I 1 0 0))
				appYm = applyIf (isYMinusIn i) (disconnectToward (V.Vec3I 0 (-1) 0))
				appYp = applyIf (isYPlusIn i) (disconnectToward (V.Vec3I 0 1 0))
				appZm = applyIf (isZMinusIn i) (disconnectToward (V.Vec3I 0 0 (-1)))
				appZp = applyIf (isZPlusIn i) (disconnectToward (V.Vec3I 0 0 1))

		applyIf cond f x
			|cond = f x
			|otherwise = x

		isXMinusIn (V.Vec3I x _ _) = x-1==x0
		isXPlusIn (V.Vec3I x _ _) = x+1==x1
		isYMinusIn (V.Vec3I _ y _) = y-1==y0
		isYPlusIn (V.Vec3I _ y _) = y+1==y1
		isZMinusIn (V.Vec3I _ _ z) = z-1==z0
		isZPlusIn (V.Vec3I _ _ z) = z+1==z1

		disconnectToward n_dis (TVoxel rs) = TVoxel $ map disconnectRegion rs
			where
				disconnectRegion (TVRegion empty faces) =
					TVRegion empty $ mapFacesWithNormal f faces
				f (n,conns)
					|n == n_dis = S.empty
					|otherwise = conns

		isBoundary (V.Vec3I x y z) =
			(x==x0 || x==x1) ||
			(y==y0 || y==y1) ||
			(z==z0 || z==z1)

		(V.Vec3I x0 y0 z0, V.Vec3I x1 y1 z1) = bounds arr

-- | Add 1-thick filled cells around given model.
-- this operation shouldn't change connections of internal cells, so just copying them is ok.
extend :: TVolume -> TVolume
extend (TVolume arr) = TVolume arr'
	where
		arr' = listArray bnd' $ map convert $ range bnd'

		convert i
			|Data.Ix.inRange bnd i = arr!i
			|otherwise = TVoxel []

		bnd' = (b0-V.Vec3I 1 1 1, b1+V.Vec3I 1 1 1)
		bnd@(b0,b1) = bounds arr

main = testDDF

testDDF = do
	[p0x,p0y,p0z, path_scene, path_in, s_ntheta, s_unit, path_out] <- getArgs
	let n_theta = read s_ntheta
	let unit = read s_unit
	let p0 = V.Vec3D (read p0x) (read p0y) (read p0z)
	tv <- liftM TraceVoxel.extend $ loadJSONTV path_in
	dumpDDFMP p0 path_scene path_out n_theta unit $ generateDDF n_theta tv

testPartialDDF = do
	[path_in] <- getArgs
	tv <- liftM fillBoundary $ loadJSONTV path_in
	dumpPartialDDF n_theta tv
	where n_theta = 6

testDDS = do
	[path] <- getArgs
	tv <- liftM fillBoundary $ loadJSONTV path
	print $ generateDDS 5 tv (V.Vec3I 5 5 5)

testDDR = do
	[path] <- getArgs
	tv <- liftM fillBoundary $ loadJSONTV path
	print =<< generateDDR 3 tv (V.Vec2I 0 0) (V.Vec3I 5 5 5)

testFill = do
	[path] <- getArgs
	tv <- loadJSONTV path

	let tv'@(TVolume arr') = fillBoundary tv
	print $ bounds arr'
	dumpSimpleFVXL "filled.fvxl" $ listArray (bounds arr') $ elems $ amap (\x->if isVoxelEmpty x then 5 else 0x80) arr'

testShaft = do
	dumpAllShaft 3
	
