# cython: profile=False, cdivision=True, boundscheck=False, wraparound=False
from __future__ import print_function, division
from cython cimport bool
from libc.stdlib cimport malloc, free
cimport numpy as np
import numpy as np
import scipy.linalg as la
from tools3d.mesh import Mesh

cdef double _raycast_internal(np.ndarray[double,ndim=3] tris, np.ndarray[double,ndim=3] bbs, double[3] fr, double[3] d):
	cdef double eps = 1e-6 # 1um
	cdef double t_min = 1e6 # 1Mm = inf
	# in-loop
	cdef double[3] p
	cdef double den, t, a, b

	for i in range(tris.shape[0]):
		# den = np.dot(d, bbs[i,2])
		den = 0
		for ix in range(3):
			den += d[ix] * bbs[i,2,ix]
		if den==0:
			continue # parallel

		# t = np.dot(tris[i,0]-fr, bbs[i,2]) / den
		t = 0
		for ix in range(3):
			t += (tris[i,0,ix]-fr[ix]) * bbs[i,2,ix]
		t /= den

		if t<0:
			continue

		# p = fr + d * t - tris[i,0]
		for ix in range(3):
			p[ix] = fr[ix] + d[ix] * t - tris[i,0,ix]

		# a = np.dot(p, bbs[i,0])
		# b = np.dot(p, bbs[i,1])
		a = 0
		b = 0
		for ix in range(3):
			a += p[ix] * bbs[i,0,ix]
			b += p[ix] * bbs[i,1,ix]

		if not (-eps<a and -eps<b and a+b<1+eps):
			continue # not inside

		# intersect!
		t_min = min(t, t_min)

	return t_min


class Accel(object):
	"""
	Reference
		http://tog.acm.org/resources/RTNews/html/rtnews5b.html#art3 Arenberg
	"""
	def __init__(self, mesh):
		self.mesh = mesh
		self.tris = np.array([[mesh.vs[i0], mesh.vs[i1], mesh.vs[i2]] for (i0,i1,i2) in mesh.fs])
		self.bbs = np.zeros([len(mesh.fs), 3, 3])

		for (ix, (v0,v1,v2)) in enumerate(self.tris):
			d1 = v1 - v0
			d2 = v2 - v0

			n = np.cross(d1, d2)
			n /= la.norm(n)

			self.bbs[ix] = la.inv(np.array([d1,d2,n]).T)

	def raycast(self, fr, d):
		""" d must be normalized """
		cdef double[3] fr_array, d_array

		for i in range(3):
			fr_array[i] = fr[i]
			d_array[i] = d[i]

		return _raycast_internal(self.tris, self.bbs, fr_array, d_array)

# you can actually use union
cdef struct BIHNode:
	int is_leaf
	# is_leaf = False
	int axis
	double vl, vh
	BIHNode *cl, *ch
	# is_leaf = True
	double tri[3][3]
	double bbs[3][3]

cdef double _raycast_internal_tri(
	double tri[3][3], double bbs[3][3],
	double[3] fr, double[3] d):

	cdef double eps = 1e-6 # 1um
	cdef double[3] p
	cdef double den, t, a, b

	# den = np.dot(d, bbs[i,2])
	den = 0
	for ix in range(3):
		den += d[ix] * bbs[2][ix]
	if den==0:
		return 1e6

	# t = np.dot(tris[i,0]-fr, bbs[i,2]) / den
	t = 0
	for ix in range(3):
		t += (tri[0][ix]-fr[ix]) * bbs[2][ix]
	t /= den

	if t<0:
		return 1e6

	# p = fr + d * t - tris[i,0]
	for ix in range(3):
		p[ix] = fr[ix] + d[ix] * t - tri[0][ix]

	# a = np.dot(p, bbs[i,0])
	# b = np.dot(p, bbs[i,1])
	a = 0
	b = 0
	for ix in range(3):
		a += p[ix] * bbs[0][ix]
		b += p[ix] * bbs[1][ix]

	if not (-eps<a and -eps<b and a+b<1+eps):
		return 1e6

	return t

cdef double _raycast_bih_internal(
	double[3] fr, double[3] d,
	double tmin, double tmax,
	BIHNode *n):
	
	cdef int axis
	cdef double tl,th
	cdef double fr_axis, d_axis

	if n.is_leaf:
		return _raycast_internal_tri(n.tri, n.bbs, fr, d)
	else:
		d_axis = d[n.axis]

		if d_axis == 0:
			return min(
				_raycast_bih_internal(fr,d,tmin,tmax,n.cl),
				_raycast_bih_internal(fr,d,tmin,tmax,n.ch))
		else:
			d_axis = 1 / d_axis
			fr_axis = fr[n.axis]
			tl = (n.vl-fr_axis) * d_axis
			th = (n.vh-fr_axis) * d_axis

			if d_axis>0: # l to h
				return min(
					_raycast_bih_internal(fr,d,tmin,min(tmax,tl),n.cl) if tmin<=tl else 1e6,
					_raycast_bih_internal(fr,d,max(tmin,th),tmax,n.ch) if tmax>=th else 1e6)
			else: # h to l
				return min(
					_raycast_bih_internal(fr,d,tmin,min(tmax,th),n.ch) if tmin<=th else 1e6,
					_raycast_bih_internal(fr,d,max(tmin,tl),tmax,n.cl) if tmax>=tl else 1e6)

cdef BIHNode *_create_bih_internal(
	np.ndarray[double,ndim=1] bmin,
	np.ndarray[double,ndim=1] bmax,
	np.ndarray[double,ndim=3] tris):

	cdef BIHNode *n
	cdef int ix_longest, ix_split, ix_pivot
	cdef int n_tris = tris.shape[0]
	cdef double center
	cdef double temp
	cdef np.ndarray[double,ndim=2] bbs

	if n_tris == 0:
		raise ValueError("impossible happened! wrong partition in BIH")
	elif n_tris == 1:
		n = <BIHNode*>malloc(sizeof(BIHNode))
		n.is_leaf = 1
		
		bbs = np.empty([3,3])

		# bbs := (v1-v0, v2-v0, NA)
		for i in range(3):
			bbs[i,0] = tris[0,1,i] - tris[0,0,i]
			bbs[i,1] = tris[0,2,i] - tris[0,0,i]

		# bbs := (_, _, cross)
		bbs[:,2] = np.cross(bbs[:,0], bbs[:,1])
		bbs[:,2] /= la.norm(bbs[:,2])

		bbs = la.inv(bbs)

		# store as fixed array
		for i in range(3):
			for j in range(3):
				n.tri[i][j] = tris[0,i,j]
				n.bbs[i][j] = bbs[i,j]

		return n
	else:
		ix_longest = (bmax-bmin).argmax()
		box_center = (bmin+bmax)/2

		ix_split = ix_longest

		# separate tris to 2 overlapping regions, L and H
		center = box_center[ix_split]

		# sort in-place: L=[:ix_pivot] H=[ix_pivot:]
		ix_pivot = 0
		for i in range(n_tris):
			if (tris[i,0,ix_split]+tris[i,1,ix_split]+tris[i,2,ix_split])/3<center:
				# swap(i,ix_pivot)
				for di in range(3):
					for dj in range(3):
						temp = tris[i,di,dj]
						tris[i,di,dj] = tris[ix_pivot,di,dj]
						tris[ix_pivot,di,dj] = temp
				ix_pivot += 1

		# very pathological case: move whatever first element to L region
		if ix_pivot == 0 or ix_pivot == n_tris:
			ix_pivot = n_tris // 2

		# calculate intervals for L and H
		vL = tris[:ix_pivot,:,ix_split].max()
		vH = tris[ix_pivot:,:,ix_split].min()

		bmaxL = bmax.copy()
		bmaxL[ix_split] = vL
		bminH = bmin.copy()
		bminH[ix_split] = vH

		#print(bmin, bmax, center, ix_split, ix_pivot, len(tris))

		n = <BIHNode*>malloc(sizeof(BIHNode))
		n.is_leaf = 0
		n.axis = ix_split
		n.vl = vL
		n.vh = vH
		n.cl = _create_bih_internal(bmin, bmaxL, tris[:ix_pivot])
		n.ch = _create_bih_internal(bminH, bmax, tris[ix_pivot:])

		return n

cdef _delete_bih_internal(BIHNode* n):
	if not n.is_leaf:
		_delete_bih_internal(n.cl)
		_delete_bih_internal(n.ch)
	free(n)


cdef class BIHAccel(object):
	""" Bounding Interval Hierarchy accelerator"""
	cdef BIHNode *tree
	cdef double tmax0
	def __init__(self, mesh):
		bmin = mesh.vs.min(axis=0)
		bmax = mesh.vs.max(axis=0)
		self.tree = _create_bih_internal(bmin, bmax, mesh.flatten())
		self.tmax0 = la.norm(bmax-bmin)

	def __del__(self):
		_delete_bih_internal(self.tree)

	def raycast(self, fr, d):
		cdef double[3] fr_array, d_array

		for i in range(3):
			fr_array[i] = fr[i]
			d_array[i] = d[i]

		return _raycast_bih_internal(fr_array, d_array, 0, self.tmax0, self.tree)

cdef class AABBAccel(object):
	""" AABB intersection accelerator """
	cdef double[3] minB, maxB
	def __init__(self, vmin, vmax):
		for i in range(3):
			self.minB[i] = vmin[i]
			self.maxB[i] = vmax[i]

	def raycast(self, fr, d):
		"""
		http://tog.acm.org/resources/GraphicsGems/gems/RayBox.c
		"""
		cdef double[3] fr_array, d_array
		cdef int RIGHT=0, LEFT=1, MIDDLE=2
		cdef int inside = True
		cdef int[3] quadrant
		cdef int whichPlane=0
		cdef double[3] maxT
		cdef double[3] candidatePlane
		cdef double tCoord

		for i in range(3):
			fr_array[i] = fr[i]
			d_array[i] = d[i]

		# find candidate planes
		for i in range(3):
			if fr_array[i] < self.minB[i]:
				quadrant[i] = LEFT
				candidatePlane[i] = self.minB[i]
				inside = False
			elif fr_array[i] > self.maxB[i]:
				quadrant[i] = RIGHT
				candidatePlane[i] = self.maxB[i]
				inside = False
			else:
				quadrant[i] = MIDDLE;

		if inside:
			return 0

		# Calculate T distances to candidate planes
		for i in range(3):
			if quadrant[i] != MIDDLE and d_array[i] != 0:
				maxT[i] = (candidatePlane[i]-fr_array[i]) / d_array[i]
			else:
				maxT[i] = -1

		# Get largest of the maxT's for final choice of intersection
		for i in range(3):
			if maxT[whichPlane] < maxT[i]:
				whichPlane = i

		# Check final candidate actually inside box
		if maxT[whichPlane] < 0:
			return 1e6

		for i in range(3):
			if whichPlane != i:
				tCoord = fr_array[i] + maxT[whichPlane] * d_array[i]
				if tCoord < self.minB[i] or tCoord > self.maxB[i]:
					return 1e6
			
		return maxT[whichPlane]