# coding=utf-8
from __future__ import print_function, division
import sys
import msgpack
import math
import cmath
import random
import numpy as np
import scipy.linalg as la
from tools3d.affine import *
from tools3d.mesh import *
from scene import Scene
import os.path
import traceback

class ErrorModel(object):
	def get_desc(self):
		return '0'

	def __call__(self, t):
		if type(t) is tuple:
			return t
		else:
			return (t,t)

class LinearErrorModel(ErrorModel):
	def __init__(self, gradient):
		self.gradient = gradient
		self.rmin = 1-self.gradient/2
		self.rmax = 1+self.gradient/2

	def get_desc(self):
		return 'linear(%fm/m)'%self.gradient

	def __call__(self, t):
		if type(t) is tuple:
			return (t[0]*self.rmin, t[1]*self.rmax)
		else:
			return (t*self.rmin, t*self.rmax)

class DefaultDeviceModel(object):
	def __init__(self):
		self.have_accel = True
		self.fovx = math.radians(58)
		self.fovy = math.radians(45)
		self.error = LinearErrorModel(0.2/4)

	def get_desc(self):
		return 'eps=%s / FOV:%d * %d'%(self.error.get_desc(), int(math.degrees(self.fovx)), int(math.degrees(self.fovy)))

	def get_field(self, aunit):
		return DiscreteDirection(aunit) # whatever will do (for visualize_index.py, at least)


class Index(object):
	def __init__(self, o):
		self.o = o

		self.grid = o['grid']
		
		self.direction = DiscreteDirection(self.grid['aunit'])
		self.orientation = DiscreteOrientation(self.grid['aunit'])

		self.shape = self.pos_shape + [self.orientation.ntheta, self.orientation.nphi, self.orientation.nalpha]

		self.npos = self.shape[0] * self.shape[1] * self.shape[2]
		self.nori = self.shape[3] * self.shape[4] * self.shape[5]

		if type(o['samples']) is tuple:
			flat_samples = np.array(o['samples'], np.float32)
		else:
			flat_samples = np.fromstring(o['samples'], np.float32)

		print(self.pos_shape)
		print(self.orientation)
		print(self.direction)
		print(len(flat_samples))

		self.samples = flat_samples.reshape([self.npos,len(self.direction),2])
	
	@staticmethod
	def load_from(path):
		return Index(msgpack.unpack(open(path,'rb')))

	@property
	def pos_shape(self):
		return [self.grid[k] for k in 'x y z'.split()]

	@property
	def rot_shape(self):
		return self.shape[3:]

	@property
	def samples_full(self):
		""" Fully indexed (x,y,z,theta,phi,min-max) sample array. """
		return self.samples.reshape(self.pos_shape + [self.direction.ntheta, self.direction.nphi, 2])

	def count_valid_cells(self):
		n_valid_pos = 0
		for ipos in xrange(self.npos):
			if (self.samples[ipos,:,0]>1e3).all():
				continue
			n_valid_pos += 1
		return n_valid_pos * self.nori

	def localize(self, dev, dmap, ori_filter=None):
		if ori_filter == None:
			ori_filter = set(xrange(len(self.orientation)))

		locs = []
		for iori in ori_filter:
			ts = dev.rotate_discretize(self.direction, dmap, self.orientation, iori)
			
			#ts[:,0] *= 1.025
			#ts[:,1] *= 0.975
			#ts[:,0] *= 0.975
			#ts[:,1] *= 1.025
			for ipos in xrange(self.npos):
				if self.compatible(self.samples[ipos], ts):
					locs.append((ipos, iori))
		return set(locs)

	def to_legacy(self, locs):
		flag = np.zeros([self.npos,self.nori], bool)
		for (ipos,iori) in locs:
			flag[ipos,iori] = True

		return flag.reshape(self.shape)

	def to_elements(self, locs):
		res = set()
		for (ipos, iori) in locs:
			tpos = np.unravel_index(ipos, tuple(self.shape[:3]))
			tori = np.unravel_index(iori, tuple(self.shape[3:]))
			res.add(tpos+tori)
		return res

	def compatible(self, dm0, dm1):
		#return (dm0[:,0] <= dm1[:,0]).all() and (dm1[:,1] <= dm0[:,1]).all() # dm0 contains dm1 for all pos
		return (dm0[:,0] < dm1[:,1]).all() and (dm0[:,1] > dm1[:,0]).all() # overlap exists for all pos


class DiscreteCameraField(object):
	"""
	Discretization of camera field of view.
	Provides iterator to represent continuous sub-region.

	camera forward is mapped to Z+

	continuous repr.: 3d unit vector as np.array
	discrete repr.: integer
	"""
	def __init__(self, fovw, fovh, max_angle):
		assert(fovw<math.pi)
		assert(fovh<math.pi)

		self.fovx = fovw
		self.fovy = fovh

		self.nx = int(self.fovx/max_angle)
		self.ny = int(self.fovy/max_angle)

		self.confusion =  (self.fovx/self.nx+self.fovy/self.ny)/2 # confusion is largest at the center

		self.sx = math.tan(self.fovx/2)*2
		self.sy = math.tan(self.fovy/2)*2

		self.shape = (self.nx, self.ny)
	
	def __len__(self):
		return self.nx * self.ny

	def sample(self, i):
		ix, iy = np.unravel_index(i, self.shape)

		dx = self.sx * ((ix+random.random())/self.nx-0.5)
		dy = self.sy * ((iy+random.random())/self.ny-0.5)

		d = np.array([dx,dy,1])
		return d/la.norm(d)

	def multi_sample(self, i, n):
		ix, iy = np.unravel_index(i, self.shape)

		ds = np.empty([n,3])
		ds[:,0] = self.sx * ((ix+np.random.rand(n))/self.nx-0.5)
		ds[:,1] = self.sy * ((iy+np.random.rand(n))/self.ny-0.5)
		ds[:,2] = 1

		return ds / np.sqrt((ds**2).sum(axis=1))[:,np.newaxis]

	def center(self, i):
		ix, iy = np.unravel_index(i, self.shape)

		dx = self.sx * ((ix+0.5)/self.nx-0.5)
		dy = self.sy * ((iy+0.5)/self.ny-0.5)

		d = np.array([dx,dy,1])
		return d/la.norm(d)

	def cone(self, i):
		return (self.center(i), self.confusion)

	def __repr__(self):
		return 'DiscreteCameraField(nx=%d, ny=%d)'%(self.nx, self.ny)

	def subset_as_2d(self, es, i):
		ix, iy = np.unravel_index(i, self.shape)



class DiscreteDirection(object):
	"""
	Discretization of 3d direction.
	Supports decomposition into 6 axis-aligned hemispheres

	continuous: 3d unit vector as np.array
	discrete repr: integer

	theta, phi region is actually [,)
	north pole -> (0,0)
	south pole -> (self.ntheta-1,0)
	"""
	def __init__(self, max_angle):
		nt = int(math.pi/max_angle) # this is not enough to satisfy max_angle
		if nt % 2 == 1:
			nt += 1

		self.ntheta = nt
		self.nphi = nt*2
		self.shape = (self.ntheta, self.nphi)

		#self.confusion = self.calc_obvious_confusion()
		self.confusion = self.calc_empirical_confusion()
		self.center_cache = [self._center(i) for i in xrange(len(self))]

	def calc_obvious_confusion(self):
		"""
		Calculate confusion radius which is obviously correct, but a bit larger than real value.
		"""
		return 2 * (math.pi/self.ntheta)/2 # delta_theta/2 + delta_phi/2

	def calc_empirical_confusion(self):
		"""
		Calculate empirical confusion radius.
		This is correct if there's no point other than vertices that maximize the confusion.
		"""
		cf = 0
		for idir in xrange(len(self)):
			nc = self.realize_fract(idir, [0.5,0.5])
			ns = [self.realize_fract(idir, d) for d in [[di,dj] for di in [0,1] for dj in [0,1]]]
			cf = max(cf, max(math.acos(np.dot(nc,n)) for n in ns))
		return cf


	def __len__(self):
		return self.ntheta * self.nphi

	def discretize(self, d):
		theta = math.atan2(la.norm(d[:2]), d[2])
		phi = math.atan2(d[1], d[0]) % (2*math.pi)

		return np.ravel_multi_index((
			min(self.ntheta-1,int(self.ntheta*theta/math.pi)),
			min(self.nphi-1, int(self.nphi*phi/(2*math.pi)))), self.shape)

	def multi_discretize(self, ds):
		rs = (ds[:,:2]**2).sum(axis=1)**0.5

		ts = np.arctan2(rs, ds[:,2])
		ps = np.arctan2(ds[:,1], ds[:,0])
		ps = np.where(ps<0, ps+2*math.pi, ps)

		its = np.floor(ts*(self.ntheta/math.pi)).astype(int)
		ips = np.floor(ps*(self.nphi/2/math.pi)).astype(int)

		return np.ravel_multi_index((np.minimum(its,self.ntheta-1), ips), self.shape)

	def empty(self, ty):
		return np.empty([len(self)], ty)

	def hemisphere(self, d):
		""" dir must be [XYZ][+-] """
		if d[0]>0:
			for itheta in xrange(self.ntheta):
				for iphi in xrange(3*self.nphi//4, self.nphi):
					yield np.ravel_multi_index((itheta,iphi), self.shape)
				for iphi in xrange(self.nphi//4):
					yield np.ravel_multi_index((itheta,iphi), self.shape)
		elif d[0]<0:
			for itheta in xrange(self.ntheta):
				for iphi in xrange(self.nphi//4, 3*self.nphi//4):
					yield np.ravel_multi_index((itheta,iphi), self.shape)
		elif d[1]>0:
			for itheta in xrange(self.ntheta):
				for iphi in xrange(self.nphi//2):
					yield np.ravel_multi_index((itheta,iphi), self.shape)
		elif d[1]<0:
			for itheta in xrange(self.ntheta):
				for iphi in xrange(self.nphi//2,self.nphi):
					yield np.ravel_multi_index((itheta,iphi), self.shape)
		elif d[2]>0:
			for itheta in xrange(0,self.ntheta//2):
				for iphi in xrange(self.nphi):
					yield np.ravel_multi_index((itheta,iphi), self.shape)
		elif d[2]<0:
			for itheta in xrange(self.ntheta//2, self.ntheta):
				for iphi in xrange(self.nphi):
					yield np.ravel_multi_index((itheta,iphi), self.shape)

	def realize_fract(self, i, f):
		itheta, iphi = np.unravel_index(i, self.shape)

		theta = math.pi * (f[0]+itheta)/self.ntheta
		phi = 2 * math.pi * (f[1]+iphi)/self.nphi

		return np.array([
			math.sin(theta)*math.cos(phi),
			math.sin(theta)*math.sin(phi),
			math.cos(theta)
			])

	def sample(self, i):
		itheta, iphi = np.unravel_index(i, self.shape)

		theta = math.pi * (random.random()+itheta)/self.ntheta
		phi = 2 * math.pi * (random.random()+iphi)/self.nphi

		return np.array([
			math.sin(theta)*math.cos(phi),
			math.sin(theta)*math.sin(phi),
			math.cos(theta)
			])

	def multi_sample(self, i, n):
		itheta, iphi = np.unravel_index(i, self.shape)

		ts = math.pi * (np.random.rand(n)+itheta)/self.ntheta
		ps = 2 * math.pi * (np.random.rand(n)+iphi)/self.nphi

		return np.array([
			np.sin(ts)*np.cos(ps),
			np.sin(ts)*np.sin(ps),
			np.cos(ts)
			]).T

	def center(self, i):
		return self.center_cache[i]

	def _center(self, i):
		itheta, iphi = np.unravel_index(i, self.shape)

		theta = math.pi * (0.5+itheta)/self.ntheta
		phi = 2 * math.pi * (0.5+iphi)/self.nphi

		return np.array([
			math.sin(theta)*math.cos(phi),
			math.sin(theta)*math.sin(phi),
			math.cos(theta)
			])

	def cone(self, i):
		return (self.center(i), self.confusion) # this can be improved by creating minimum bounding circle of the spherical angle

	def cones(self):
		""" Return (dirs [N,3], radius [1]). """
		return (np.array(self.center_cache, np.float32), self.confusion)

	def __repr__(self):
		return 'DiscreteDirection(ntheta=%d, nphi=%d)'%(self.ntheta, self.nphi)


class DiscreteOrientation(object):
	"""
	Discretization of SO(3).
	"""
	def __init__(self, max_angle):
		self.ntheta = int(math.pi/max_angle)
		self.nphi = 2 * self.ntheta
		self.nalpha = 2 * self.ntheta
		self.shape = (self.ntheta, self.nphi, self.nalpha)

		self.confusion = self.calc_empirical_confusion()

	def __len__(self):
		return self.ntheta * self.nphi * self.nalpha

	def calc_obvious_confusion(self):
		return (math.pi/self.ntheta) * 3 / 2 # worst case: 3 axes

	def calc_empirical_confusion(self):
		def angle(mat):
			return cmath.phase(la.eig(mat)[0][0])

		cf = 0
		for iori in xrange(len(self)):
			center = self.realize_as_matrix(iori, [0.5,0.5,0.5])
			inv_center = la.inv(center)
			vs = [self.realize_as_matrix(iori, d) for d in [[di,dj,dk] for di in [0,1] for dj in [0,1] for dk in [0,1]]]
			cf = max(cf, max(angle(np.dot(v, inv_center)) for v in vs))
		return cf

	def sample(self, i):
		it, ip, ia = np.unravel_index(i, self.shape)

		theta = math.pi * (it+random.random())/self.ntheta
		phi = 2 * math.pi * (ip+random.random())/self.nphi
		alpha = 2 * math.pi * (ia+random.random())/self.nalpha

		return (
			SO3.rotate(np.array([0,0,1]), phi) *
			SO3.rotate(np.array([0,1,0]), theta) *
			SO3.rotate(np.array([0,0,1]), alpha))


	def realize_as_matrix(self, i, frac):
		it, ip, ia = np.unravel_index(i, self.shape)

		theta = math.pi * (it+frac[0])/self.ntheta
		phi = 2 * math.pi * (ip+frac[1])/self.nphi
		alpha = 2 * math.pi * (ia+frac[2])/self.nalpha

		def yrot(c, s):
			return np.array([
				[ c,0,s],
				[ 0,1,0],
				[-s,0,c]])

		def zrot(c, s):
			return np.array([
				[c,-s,0],
				[s, c,0],
				[0, 0,1]])

		return np.dot(zrot(math.cos(phi), math.sin(phi)),
					np.dot(yrot(math.cos(theta), math.sin(theta)),
						zrot(math.cos(alpha), math.sin(alpha))))

	def sample_as_matrix(self, i):
		it, ip, ia = np.unravel_index(i, self.shape)

		theta = math.pi * (it+random.random())/self.ntheta
		phi = 2 * math.pi * (ip+random.random())/self.nphi
		alpha = 2 * math.pi * (ia+random.random())/self.nalpha

		def yrot(c, s):
			return np.array([
				[ c,0,s],
				[ 0,1,0],
				[-s,0,c]])

		def zrot(c, s):
			return np.array([
				[c,-s,0],
				[s, c,0],
				[0, 0,1]])

		return np.dot(zrot(math.cos(phi), math.sin(phi)),
					np.dot(yrot(math.cos(theta), math.sin(theta)),
						zrot(math.cos(alpha), math.sin(alpha))))

	def multi_sample_as_matrix(self, i, n):
		it, ip, ia = np.unravel_index(i, self.shape)

		theta = math.pi * (it+np.random.rand(n))/self.ntheta
		phi = 2 * math.pi * (ip+np.random.rand(n))/self.nphi
		alpha = 2 * math.pi * (ia+np.random.rand(n))/self.nalpha

		def dot_ms_ms(ms0, ms1):
			return np.einsum('sij,sjk->sik', ms0, ms1)

		def yrot(c, s):
			ms = np.zeros([n,3,3])
			ms[:,0,0] = c
			ms[:,0,2] = s
			ms[:,2,0] = -s
			ms[:,2,2] = c
			ms[:,1,1] = 1
			return ms

		def zrot(c, s):
			ms = np.zeros([n,3,3])
			ms[:,0,0] = c
			ms[:,0,1] = -s
			ms[:,1,0] = s
			ms[:,1,1] = c
			ms[:,2,2] = 1
			return ms

		return dot_ms_ms(zrot(np.cos(phi), np.sin(phi)),
					dot_ms_ms(yrot(np.cos(theta), np.sin(theta)),
						zrot(np.cos(alpha), np.sin(alpha))))

	def cone(self, i):
		"""
		Return (center orientation, confusion radius).

		orientation = <small ori> * <center ori>
		|small ori|<confusion radius

		Maximum angle change is eqaul to rotation angle.
		"""
		it, ip, ia = np.unravel_index(i, self.shape)

		theta = math.pi * (it+0.5)/self.ntheta
		phi = 2 * math.pi * (ip+0.5)/self.nphi
		alpha = 2 * math.pi * (ia+0.5)/self.nalpha

		def yrot(c, s):
			return np.array([
				[ c,0,s],
				[ 0,1,0],
				[-s,0,c]])

		def zrot(c, s):
			return np.array([
				[c,-s,0],
				[s, c,0],
				[0, 0,1]])

		rot = np.dot(zrot(math.cos(phi), math.sin(phi)),
					np.dot(yrot(math.cos(theta), math.sin(theta)),
						zrot(math.cos(alpha), math.sin(alpha))))


		return (rot, self.confusion)

	def __repr__(self):
		return 'DiscreteOrientation(ntheta=%d, nphi=%d, nalpha=%d)'%(
			self.ntheta, self.nphi, self.nalpha)
