#!/bin/env python
# coding=utf-8
from __future__ import division, print_function
import cairo
import msgpack
import numpy as np
import scipy.linalg as la
import math
import sys
from vtk import *
import vtk.util.numpy_support as vn
import numpy as np
import pyximport
pyximport.install()
import accel
import time
import types
import random
from tools3d.mesh import Mesh
from tools3d.affine import *

class DepthRenderWindow(object):
	"""
	A dynamic image of current depth map seen from the view.
	Upon view update, this class renders the image in the background,
	but always reacts to get_image() instantly by returning incomplete render.
	"""
	def __init__(self, mesh):
		self.mesh = mesh
		self.accel = accel.Accel(self.mesh)
		self.image = (np.random.rand(100,100,3)*255).astype(np.uint8)

		self.render_cr = self.render()
		self.notify_view(SE3.identity())

	def notify_view(self, c2w):
		self.c2w = c2w
		self.changed = True

	def get_image(self):
		""" This function will return current image and also drives rendering coroutine """
		self.render_cr.next()
		return self.image

	def render(self):
		# 100px : 10ms
		pv = np.zeros(3,np.uint8)

		def calc_at(i,j):
			dc = np.array([(i-50)*0.01,(j-50)*0.01,1])
			dc /= la.norm(dc)
			theta = math.degrees(math.acos(dc[2]))

			dw = self.c2w.extract_rotation() * dc

			t = self.accel.raycast(self.c2w * np.zeros(3), dw)
			pv[:] = t*100
			if int(theta)%10==0:
				pv[0] *= 0.5

		while True:
			self.changed = False

			# 1/10
			for i in xrange(10):
				for j in xrange(10):
					calc_at(i*10+5, j*10+5)
					self.image[i*10:(i+1)*10,j*10:(j+1)*10,:] = pv

			yield

			# 1/1
			ds = [(di,dj) for di in xrange(10) for dj in xrange(10)]
			random.shuffle(ds)
			for (di,dj) in ds:
					if self.changed:
						break

					# finish one pixel for each tile
					for ti in xrange(10):
						for tj in xrange(10):
							i = ti*10+di
							j = tj*10+dj
							calc_at(i,j)
							self.image[i,j,:] = pv

					yield

			# 2/1
			for i in xrange(100):
				if self.changed:
					break

				for j in xrange(100):
					ss = np.zeros(3, np.float)
					calc_at(i-0.25, j-0.25)
					ss += pv
					calc_at(i-0.25, j+0.25)
					ss += pv
					calc_at(i+0.25, j-0.25)
					ss += pv
					calc_at(i+0.25, j+0.25)
					ss += pv

					self.image[i,j] = ss/4
				yield

			while not self.changed:
				yield


class World(object):
	"""
	A VTK window that contains interactive visualization of room structure and samples.
	"""
	def __init__(self, ranges):
		self.grid = ranges['grid']
		nx = self.grid['x']
		ny = self.grid['y']
		nz = self.grid['z']
		nth = self.grid['theta']
		nph = self.grid['phi']
		nal = self.grid['alpha']
		self.samples = [np.array(s['ranges']).reshape([nx,ny,nz,nth,nph,-1,2]) for s in ranges['samples']]
		self.omegas = [s['offset-omega'] for s in ranges['samples']]

		self.mesh = Mesh.unpack_headerless(ranges['mesh'])

		#a renderer and render window
		self.renderer = vtkRenderer()
		self.window = vtkRenderWindow()
		self.window.AddRenderer(self.renderer)
		self.renderer.SetBackground(.1,.2,.3) # dark blue background

		ceiling_light = vtkLight()
		ceiling_light.SetPosition(0, 0, 1)
		ceiling_light.SetAmbientColor(0.3,0.3,0.3)
		self.renderer.AddLight(ceiling_light)

		self.drw = DepthRenderWindow(self.mesh)

		# add interactor
		self.interactor = vtkRenderWindowInteractor() 
		style = vtkInteractorStyleTrackballCamera()
		self.interactor.SetInteractorStyle(style)
		self.interactor.SetRenderWindow(self.window)

		# prepare model
		self.put_model()
		self.put_arrows()
		self.put_sliders() # must be called before put_view
		self.put_view()
		self.put_floor()
		self.put_raster()

		# attach view controller

		# put this before CreateRepeatingTimer otherwise it segfaults
		# http://1045678.n5.nabble.com/iren-gt-CreateRepeatingTimer-segfaults-td1229012.html
		self.interactor.Initialize()

		self.interactor.AddObserver('TimerEvent', self.on_timer)
		self.interactor.CreateRepeatingTimer(100)
		self.interactor.AddObserver('KeyPressEvent', self.on_keypress)

		# ins param
		self.update_arrows([1]*len(self.samples))

	def on_timer(self, obj, ev):
		self.update_raster()
		self.window.Render()

	def on_keypress(self, obj, ev):
		table = {
			# translation
			'Up': SE3.translate(0.1*np.array([0,1,0])),
			'Down': SE3.translate(0.1*np.array([0,-1,0])),
			'Left': SE3.translate(0.1*np.array([-1,0,0])),
			'Right': SE3.translate(0.1*np.array([1,0,0])),
			'space': SE3.translate(0.1*np.array([0,0,1])),
			'Shift_L': SE3.translate(0.1*np.array([0,0,-1])),
			# rotation
			'Control_L': SE3.rotate(np.array([1,0,0]), math.pi*0.5),
			'Alt_L': SE3.rotate(np.array([0,1,0]), math.pi*0.5),
			'Shift_R': SE3.rotate(np.array([0,0,1]), math.pi*0.5)
		}

		k = obj.GetKeySym()
		if k in table:
			self.cam2world = self.cam2world * table[k]

			self.drw.notify_view(self.cam2world)
			self.update_view()
			self.window.Render()
		else:
			print('unhandled',k)

	def run(self):
		self.renderer.ResetCamera()
		self.window.Render()
		self.interactor.Start()

	def put_view(self):
		r = 0.2
		self.cam2world = SE3.translate(np.array([0.5,0.5,0.5])) * SE3.rotate(np.array([1,0,0]), math.radians(90))

		# local dir sphere
		self.view_rings = []
		for phi0 in [10*ix for ix in range(4)]:
			phi1 = phi0+1
			src = vtkSphereSource()
			src.SetRadius(r)
			src.SetStartPhi(phi0)
			src.SetEndPhi(phi1)
			src.SetPhiResolution(5)
			src.SetThetaResolution(30)

			transformFilter = vtkTransformPolyDataFilter()
			transformFilter.SetTransform(vtkTransform())
			transformFilter.SetInputConnection(src.GetOutputPort())

			m = vtkPolyDataMapper()
			m.SetInputConnection(transformFilter.GetOutputPort())

			a = vtkActor()
			a.SetMapper(m)
			a.GetProperty().SetAmbient(0.1)

			self.renderer.AddActor(a)

			self.view_rings.append(transformFilter)

		# delta omegas
		self.view_directions = []
		for om in self.omegas:
			# ray
			src_ray = vtkLineSource()

			m = vtkPolyDataMapper()
			m.SetInput(src_ray.GetOutput())
			 
			a = vtkActor()
			a.SetMapper(m)
			a.GetProperty().SetColor(0.5,0.9,0.8)

			self.renderer.AddActor(a)

			# sphere
			src_sphere = vtkSphereSource()
			src_sphere.SetRadius(0.01)

			m = vtkPolyDataMapper()
			m.SetInput(src_sphere.GetOutput())

			a = vtkActor()
			a.SetMapper(m)
			a.GetProperty().SetColor(0.5,0.8,0.5)
			a.GetProperty().SetAmbient(0.1)

			self.renderer.AddActor(a)

			#
			self.view_directions.append({
				'ray': src_ray,
				'sphere': src_sphere
				})

		self.update_view()

	def update_view(self):
		r = 0.2

		# reset view
		for ring in self.view_rings:
			ring.SetTransform(self.cam2world.as_vtk())

		# cast rays
		acc = accel.Accel(self.mesh)
		o = self.cam2world * np.zeros(3)
		ts = []
		for (om, sl, vd) in zip(self.omegas, self.sliders, self.view_directions):
			dc = np.array([
				math.sin(om['theta'])*math.cos(om['phi']),
				math.sin(om['theta'])*math.sin(om['phi']),
				math.cos(om['theta'])])

			dw = self.cam2world.extract_rotation() * dc

			# set view origin
			vd['sphere'].SetCenter(self.cam2world * (r*dc))
			vd['ray'].SetPoint1(self.cam2world * np.zeros(3))

			if sl.GetRepresentation().GetValue()<0.01:
				ts.append(None)
			else:
				try:
					t = acc.raycast(o, dw)
					ts.append(t)

					sl.GetRepresentation().SetValue(t)
					vd['ray'].SetPoint2(self.cam2world * (t*dc))
				except ValueError:
					ts.append(None)

					sl.GetRepresentation().SetValue(3)
					vd['ray'].SetPoint2(self.cam2world * (3*dc))

		self.update_arrows(ts)

	def put_sliders(self):
		"""
		TODO:
		show omegas visually, or making omega selectable at run-time is even better
		"""
		n = len(self.samples)

		def cb(obj, ev):
			ds = []
			for i in range(n):
				rep = self.sliders[i].GetRepresentation()
				d = rep.GetValue()

				if d < 0.01:
					rep.GetLabelProperty().SetOpacity(0)
					rep.GetTubeProperty().SetColor(0.2,0.2,0.2)
					rep.GetCapProperty().SetColor(0.2,0.2,0.2)
					ds.append(None)
				else:
					rep.GetLabelProperty().SetOpacity(1)
					rep.GetTubeProperty().SetColor(1,1,1)
					rep.GetCapProperty().SetColor(1,1,1)
					ds.append(d)

			try:
				self.update_arrows(ds)
			except:
				import traceback
				traceback.print_exc()

			self.update_view()

		self.sliders = []
		for (i, omega) in enumerate(self.omegas):
			rep = vtkSliderRepresentation2D()
			rep.SetTitleText('theta=%d, phi=%d'%(math.degrees(omega['theta']), math.degrees(omega['phi'])))
			rep.SetMinimumValue(0)
			rep.SetMaximumValue(3)
			rep.SetValue(1)
			rep.GetPoint1Coordinate().SetCoordinateSystemToNormalizedDisplay()
			rep.GetPoint2Coordinate().SetCoordinateSystemToNormalizedDisplay()
			rep.GetPoint1Coordinate().SetValue(0.85, 0.05+0.05*i)
			rep.GetPoint2Coordinate().SetValue(0.98, 0.05+0.05*i)
			rep.SetTubeWidth(0.003)
			rep.SetEndCapWidth(0.02)
			rep.SetEndCapLength(0.001)
			rep.SetLabelHeight(0.02)
			rep.SetTitleHeight(0.015)

			widget = vtkSliderWidget()
			widget.SetRepresentation(rep)
			widget.SetInteractor(self.interactor)
			widget.AddObserver("InteractionEvent", cb)
			widget.On()

			self.sliders.append(widget) # w/o this, widget will be garbage collected and gone

	def put_model(self):
		# points
		pts = vtkPoints()
		for v in self.mesh.vs:
			pts.InsertNextPoint(v)

		# map each face to 3 lines
		lines = vtkCellArray()
		for (i,j,k) in self.mesh.fs:
			for (s,t) in [(i,j),(j,k),(k,i)]:
				line = vtkLine()
				line.GetPointIds().SetId(0,s)
				line.GetPointIds().SetId(1,t)
				lines.InsertNextCell(line)

		# create polydata
		wireframe = vtkPolyData()
		wireframe.SetPoints(pts)
		wireframe.SetLines(lines)

		# visualize
		mapper = vtkPolyDataMapper()
		mapper.SetInput(wireframe)

		actor = vtkActor()
		actor.SetMapper(mapper)
		actor.GetProperty().SetColor(0.5,0.5,0.5)

		self.renderer.AddActor(actor)

	def put_floor(self):
		"""
		Add floor grid
		"""
		unit = self.grid['unit']
		origin = np.array(self.grid['origin'])

		# floor tiles
		ix_pt = 0
		pts = vtkPoints()
		lines = vtkCellArray()
		for i in range(self.grid['x']):
			for j in range(self.grid['y']):
				v0 = origin + np.array([i+0,j+0,0])*unit
				v1 = origin + np.array([i+1,j+0,0])*unit
				v2 = origin + np.array([i+1,j+1,0])*unit
				v3 = origin + np.array([i+0,j+1,0])*unit
				
				pts.InsertNextPoint(v0)
				pts.InsertNextPoint(v1)
				pts.InsertNextPoint(v2)
				pts.InsertNextPoint(v3)

				for di in range(4):
					line = vtkLine()
					line.GetPointIds().SetId(0,ix_pt+di)
					line.GetPointIds().SetId(1,ix_pt+(di+1)%4)
					lines.InsertNextCell(line)

				ix_pt += 4

		# create polydata
		wireframe = vtkPolyData()
		wireframe.SetPoints(pts)
		wireframe.SetLines(lines)

		# visualize
		mapper = vtkPolyDataMapper()
		mapper.SetInput(wireframe)

		actor = vtkActor()
		actor.SetMapper(mapper)

		self.renderer.AddActor(actor)

	def put_arrows(self):
		unit = self.grid['unit']
		origin = np.array(self.grid['origin'])

		self.arrows_trans = {}
		for ix in range(self.grid['x']):
			for iy in range(self.grid['y']):
				for iz in range(self.grid['z']):
					cell_center = (np.array([ix,iy,iz])+0.5)*unit+origin

					arrowSource = vtkArrowSource()

					transform = vtkTransform()
					transform.Translate(*cell_center)
					transform.Scale(unit*0.5,0.03,0.03)

					transformFilter = vtkTransformPolyDataFilter()
					transformFilter.SetTransform(transform)
					transformFilter.SetInputConnection(arrowSource.GetOutputPort())
					transformFilter.Update()
					 
					amapper = vtkPolyDataMapper()
					amapper.SetInputConnection(transformFilter.GetOutputPort())

					actor = vtkActor()
					actor.SetMapper(amapper)
					actor.GetProperty().SetColor(1,0.8,0.5)
					actor.GetProperty().SetAmbient(0.1)

					self.renderer.AddActor(actor)
					self.arrows_trans[(ix,iy,iz)] = transformFilter

	def put_raster(self):
		img = vtkImageImport()
		img.SetDataScalarTypeToUnsignedChar()
		img.SetNumberOfScalarComponents(3)
		img.SetDataExtent(0,99, 0,99, 0,0)
		img.SetWholeExtent(0,99, 0,99, 0,0)
		self.raster = img
		self.update_raster()

		im = vtkImageMapper()
		im.SetInput(img.GetOutput()) # .GetOutput())
		im.SetColorLevel(255/2)
		im.SetColorWindow(255)
		im.SetZSlice(0)

		a = vtkActor2D()
		a.SetMapper(im)

		self.renderer.AddActor(a)

	def update_raster(self):
		st = self.drw.get_image().tostring()
		self.raster.CopyImportVoidPointer(st, len(st))

	def update_arrows(self, ds):
		"""
		TODO: actually, cone representing uncertainty is better than an arrow.
		"""

		def to_ortho(theta, phi):
			d = np.array([
					math.sin(theta)*math.cos(phi),
					math.sin(theta)*math.sin(phi),
					math.cos(theta)])
			return d

		def from_ortho(d):
			theta = math.degrees(math.atan2(la.norm(d[:2]), d[2]))
			phi = math.degrees(math.atan2(d[1], d[0]))
			return (theta, phi)

		# expand
		nth = self.grid['theta']
		nph = self.grid['phi']
		nal = self.grid['alpha']

		# construct
		unit = self.grid['unit']
		origin = np.array(self.grid['origin'])

		# pre-calculate vectors in each theta, phi bin
		dir_per_tp = np.zeros([nth,nph,3])
		for it in range(nth):
			theta = math.pi*(it+0.5)/nth
			for ip in range(nph):
				phi = 2*math.pi*(ip+0.5)/nph
				dir_per_tp[it,ip] = to_ortho(theta, phi)

		within_bound = np.empty([nth,nph,nal], bool)
		for ((ix,iy,iz), transF) in self.arrows_trans.iteritems():
			cell_center = (np.array([ix,iy,iz])+0.5)*unit+origin

			# check range cond
			within_bound[:] = True
			for (d,s) in zip(ds,self.samples):
				if d != None:
					within_bound &= (s[ix,iy,iz, :,:,:, 0]<=d)
					within_bound &= (d<=s[ix,iy,iz, :,:,:, 1])

			cnt_per_tp = within_bound.sum(axis=2)

			# calculate avg. direction
			cnt_dir = cnt_per_tp.sum()
			sum_dir = (cnt_per_tp[:,:,np.newaxis] * dir_per_tp).reshape([-1,3]).sum(axis=0)

			# avg direction
			transform = vtkTransform()
			transform.Translate(*cell_center)
			if cnt_dir > 0: # smaller cnt_dir is good
				p = cnt_dir/(nal*nth*nph)

				theta, phi = from_ortho(sum_dir)
				transform.RotateWXYZ(phi, 0,0,1)
				transform.RotateWXYZ(theta-90, 0,1,0) # point to somewhere in XZ plane
				transform.Scale(unit*0.5*(1-p), unit*0.5*(p*5+0.5), unit*0.5*(p*5+0.5))
			else:
				transform.Scale(0.001,0.001,0.001)
			transF.SetTransform(transform)



if __name__ == '__main__':
	# add vtkTransform IO to SE3
	def as_vtk(self):
		tr = vtkTransform()
		tr.SetMatrix(self.to_matrix().flatten())
		return tr

	SE3.as_vtk = types.MethodType(as_vtk, None, SE3)

	# normal code
	if len(sys.argv)==2:
		ifile = sys.argv[1]
		World(msgpack.unpack(open(ifile, 'rb'))).run()
	else:
		print('usage: interact.py [in.mp]')
		sys.exit(1)
