#!/bin/env python
# coding=utf-8
from __future__ import division, print_function
import cairo
import msgpack
import numpy as np
import scipy.linalg as la
import math
import sys
from vtk import *
import vtk.util.numpy_support as vn
import numpy as np
import pyximport
pyximport.install()
import accel
import time
import types
import random
from tools3d.mesh import Mesh
from tools3d.affine import *

class World(object):
	"""
	A VTK window that contains interactive visualization of solid angle.
	"""
	def __init__(self, path):
		self.scene_path = path
		self.scene = msgpack.unpack(open(path,'rb'))
		
		self.floor_proxy = Mesh.load('./prim/cube.mp').apply_cont(lambda vs:vs*[20,20,0.01])

		#a renderer and render window
		self.renderer = vtkRenderer()
		self.window = vtkRenderWindow()
		self.window.AddRenderer(self.renderer)
		self.renderer.SetBackground(.1,.1,.1) # dark gray background

		ceiling_light = vtkLight()
		ceiling_light.SetPosition(0, 0, 100)
		ceiling_light.SetAmbientColor(0.3,0.3,0.3)
		self.renderer.AddLight(ceiling_light)

		# setup avatar location
		self.head2avatar = SE3.translate([0,0,1.5])
		self.avatar2world = SE3.identity() # foot

		# add interactor
		self.interactor = vtkRenderWindowInteractor() 
		self.interactor.SetRenderWindow(self.window)

		for name in ['KeyPressEvent',
			'LeftButtonPressEvent', 'LeftButtonReleaseEvent', 'RightButtonPressEvent', 'RightButtonReleaseEvent']:
			self.interactor.RemoveObservers(name)

		self.view_move = None
		self.verts = []
		self.interactor.AddObserver('KeyPressEvent', self.on_key_press)
		self.interactor.AddObserver('MouseMoveEvent', self.on_mouse_move)
		self.interactor.AddObserver('LeftButtonPressEvent', self.on_leftbutton_press)
		self.interactor.AddObserver('LeftButtonReleaseEvent', self.on_leftbutton_release)
		self.interactor.AddObserver('RightButtonPressEvent', self.on_rightbutton_press)
		self.interactor.AddObserver('RightButtonReleaseEvent', self.on_rightbutton_release)

		# prepare model
		self.put_cursor()
		self.put_floor()
		self.put_floor_proxy()
		self.pm_proc = self.put_model()

		self.update_camera()

		# put this before CreateRepeatingTimer otherwise it segfaults
		# http://1045678.n5.nabble.com/iren-gt-CreateRepeatingTimer-segfaults-td1229012.html
		self.interactor.Initialize()

		self.interactor.AddObserver('TimerEvent', self.on_timer)
		self.interactor.CreateRepeatingTimer(100)


	def on_timer(self, obj, ev):
		if self.pm_proc != None:
			t0 = time.time()
			while time.time()-t0<0.05:
				try:
					self.pm_proc.next()
				except StopIteration:
					self.pm_proc = None
					break
			print(1e3*(time.time()-t0),'ms')

			self.window.Render()

	def on_key_press(self, obj, ev):
		table_avatar = {
			# translation
			'Up': SE3.translate(0.1*np.array([0,1,0])),
			'Down': SE3.translate(0.1*np.array([0,-1,0])),
			'Left': SE3.translate(0.1*np.array([-1,0,0])),
			'Right': SE3.translate(0.1*np.array([1,0,0])),
			'space': SE3.translate(0.1*np.array([0,0,1])),
			'Shift_L': SE3.translate(0.1*np.array([0,0,-1]))
		}

		k = obj.GetKeySym()
		if k in table_avatar:
			self.avatar2world = self.avatar2world * table_avatar[k]
			self.update_camera()
		elif k == 'z':
			print('refreshing voxels')
			self.gen_region()
		elif k == 'q':
			import sys
			sys.exit(0)
		elif k == 'x':
			print('current pointer pos')
			print(self.get_pointer_pos(obj.GetEventPosition()))
		else:
			print('unhandled',k)


	def on_leftbutton_press(self, obj, ev):
		self.view_move = {
			'initial': self.avatar2world,
			'start': np.array(obj.GetEventPosition()),
			'prev': np.array(obj.GetEventPosition())
		}

	def on_leftbutton_release(self, obj, ev):
		self.view_move = None

	def on_rightbutton_press(self, obj, ev):
		self.verts.append(self.get_pointer_pos(obj.GetEventPosition()))
		
		if len(self.verts)>=4:
			r_mesh = Mesh(self.verts[:4], [[0,1,3],[2,3,1]])
			actor = self.mesh_to_actor(r_mesh)
			actor.GetProperty().SetColor(1,1,1)
			self.renderer.AddActor(actor)

			self.scene['proxy_floor'] = list(self.scene.get('proxy_floor', []))
			self.scene['proxy_floor'].append(r_mesh.pack_headerless())
			
			self.commit_scene()

			self.verts = []

	def commit_scene(self):
		print('!!! WRITING to %s !!!'%self.scene_path)
		msgpack.pack(self.scene, open(self.scene_path, 'wb'))

	def on_rightbutton_release(self, obj ,ev):
		pass

	def on_mouse_move(self, obj, ev):
		# update 3d pointer
		self.update_pointer(obj.GetEventPosition())

		# handle view dragging
		if self.view_move:
			current = np.array(obj.GetEventPosition())
			d = current - self.view_move['prev']

			self.avatar2world = self.avatar2world * SE3.rotate(np.array([0,0,1]), math.pi*d[0]/1000)
			self.head2avatar = self.head2avatar * SE3.rotate(np.array([-1,0,0]), math.pi*d[1]/1000)
			self.update_camera()

			self.view_move['prev'] = current

	def update_pointer(self, pos_cursor):
		self.cursor.SetCenter(self.get_pointer_pos(pos_cursor))
		self.window.Render()

	def get_pointer_pos(self, pos_cursor):
		head2world = self.avatar2world * self.head2avatar

		size = np.array(self.window.GetSize())
		d = np.array(pos_cursor) - size/2

		k = math.tan(math.radians(90/2))/size[1]*2
		d3 = np.array([d[0]*k, 1, d[1]*k])

		ray_o = head2world * np.zeros(3)
		ray_d = head2world.extract_rotation() * d3

		# intersect with scene
		acc = accel.Accel(self.floor_proxy)
		t = acc.raycast(ray_o, ray_d)
		return ray_o + t*ray_d

	def update_camera(self):
		"""
		Forward: Y+
		Up: Z+

		"""
		head2world = self.avatar2world * self.head2avatar

		cam = self.renderer.GetActiveCamera()
		cam.SetPosition(head2world * np.zeros(3))
		cam.SetFocalPoint(head2world * np.array([0,1,0]))
		cam.SetViewUp(head2world.extract_rotation() * np.array([0,0,1]))
		cam.SetViewAngle(90) # vertical fov
		cam.SetClippingRange(0.1,100)
		self.window.Render()

	def run(self):
		self.renderer.ResetCamera()
		self.window.Render()
		self.interactor.Start()

	def put_cursor(self):
		src = vtkSphereSource()
		src.SetRadius(0.05)
		src.SetPhiResolution(10)
		src.SetThetaResolution(10)
		 
		m = vtkPolyDataMapper()
		m.SetInputConnection(src.GetOutputPort())

		actor = vtkActor()
		actor.SetMapper(m)
		actor.GetProperty().SetColor(0.9,0.9,0.9)
		actor.GetProperty().SetAmbient(0.1)

		self.renderer.AddActor(actor)

		self.cursor = src

	def put_floor_proxy(self):
		for fl_mesh in self.scene.get('proxy_floor', []):
			actor = self.mesh_to_actor(Mesh.unpack_headerless(fl_mesh))
			actor.GetProperty().SetColor(0.7,0.7,0.9)
			self.renderer.AddActor(actor)

	def put_floor(self):
		"""
		Add 100m by 100m floor grid
		"""
		ix_pt = 0
		pts = vtkPoints()
		lines = vtkCellArray()
		colors = vtk.vtkUnsignedCharArray()
		colors.SetNumberOfComponents(3)
		colors.SetName("colors")

		for i in range(-50,51):
			v0 = np.array([i,-50,0])
			v1 = np.array([i, 50,0])
			pts.InsertNextPoint(v0)
			pts.InsertNextPoint(v1)
			colors.InsertNextTuple3(100,255,100)
			colors.InsertNextTuple3(100,255,100)

			line = vtkLine()
			line.GetPointIds().SetId(0,ix_pt+0)
			line.GetPointIds().SetId(1,ix_pt+1)
			lines.InsertNextCell(line)

			ix_pt += 2

		for j in range(-50,51):
			v0 = np.array([-50,j,0])
			v1 = np.array([ 50,j,0])
			pts.InsertNextPoint(v0)
			pts.InsertNextPoint(v1)
			colors.InsertNextTuple3(255,100,100)
			colors.InsertNextTuple3(255,100,100)

			line = vtkLine()
			line.GetPointIds().SetId(0,ix_pt+0)
			line.GetPointIds().SetId(1,ix_pt+1)
			lines.InsertNextCell(line)

			ix_pt += 2

		# create polydata
		wireframe = vtkPolyData()
		wireframe.SetPoints(pts)
		wireframe.SetLines(lines)
		wireframe.GetPointData().SetScalars(colors)

		# visualize
		mapper = vtkPolyDataMapper()
		mapper.SetInput(wireframe)

		actor = vtkActor()
		actor.SetMapper(mapper)

		self.renderer.AddActor(actor)

	def put_model(self):
		n = 0
		for obj in self.scene['objects']:
			print('loading',obj['name'])
			mesh = Mesh.unpack_headerless(obj['mesh'])

			v0 = mesh.vs.min(axis=0)
			v1 = mesh.vs.max(axis=0)
			dv = v1-v0
			cv = (v1+v0)/2

			aabb = aabb_mesh(v0,v1)

			# replace complex small model with AABB
			if False: # la.norm(dv)<5: # len(mesh)/la.norm(dv) > 1000:
				if la.norm(dv)<0.2 and random.random()<0.9: # skip small elements randomly
					continue

				mesh = aabb

			actor = self.mesh_to_actor(mesh)
			actor.GetProperty().SetColor(0.5,0.5,0.5)
			self.renderer.AddActor(actor)

			n += 1

			yield

	def mesh_to_actor(self, mesh, as_wireframe=True):
		"""
		Convert given mesh to an vtkActor.

		cf. Fast way to convert numpy array to vtk objects:
		http://vtk.1045678.n5.nabble.com/numpy-to-vtk-and-vtkPolyData-td4949286.html
		"""
		poly = vtkPolyData()

		# points
		pts = vtkPoints()
		pts.SetData(vn.numpy_to_vtk(mesh.vs))
		poly.SetPoints(pts)

		# map each face to 3 lines
		if as_wireframe:
			lines_raw = np.empty([len(mesh.fs)*3, 3], np.int64)
			lines_raw[:,0] = 2
			lines_raw[0::3,1:] = mesh.fs[:,:2]
			lines_raw[1::3,1:] = mesh.fs[:,1:]
			lines_raw[2::3,1:] = mesh.fs[:,::-2]

			lines = vtkCellArray()
			lines.SetCells(len(lines_raw), vn.numpy_to_vtkIdTypeArray(lines_raw))
			poly.SetLines(lines)

			poly.dont_collect = [mesh.vs, lines_raw]
		else:
			tris_raw = np.empty([len(mesh.fs), 4], np.int64)
			tris_raw[:,0] = 3
			tris_raw[:,1:] = mesh.fs

			tris = vtkCellArray()
			tris.SetCells(len(tris_raw), vn.numpy_to_vtkIdTypeArray(tris_raw))
			poly.SetPolys(tris)

			poly.dont_collect = [mesh.vs, tris_raw]

		# wrap as an Actor
		mapper = vtkPolyDataMapper()
		mapper.SetInput(poly)

		actor = vtkActor()
		actor.SetMapper(mapper)

		return actor

	def gen_region(self):
		"""
		Use self.scene to generate empty voxels.
		
		Generated voxel is
		1. strictrly contained in floor proxy
		2. does not intersect with mesh
		"""
		unit = 0.5
		access_height = 2.5 # accesible height from floor plane

		vs = np.vstack([Mesh.unpack_headerless(m).vs for m in self.scene['proxy_floor']])
		aabb_min = vs.min(axis=0)
		aabb_max = vs.max(axis=0)

		ixmin = np.floor(aabb_min/unit).astype(np.int)
		ixmax = np.ceil(aabb_max/unit).astype(np.int)

		# calculate AABB for each object
		aabbs_scene = []
		for obj in self.scene['objects']:
			m = Mesh.unpack_headerless(obj['mesh'])
			aabb = (m.vs.min(axis=0), m.vs.max(axis=0))

			print(obj['name'],len(m))
			aabbs_scene.append((aabb,m.flatten()))

		# create pillars
		voxels = np.zeros([ixmax[0]-ixmin[0], ixmax[1]-ixmin[1], 10-ixmin[2]], bool)
		for ix in range(ixmin[0],ixmax[0]):
			for iy in range(ixmin[1],ixmax[1]):
				print('@%d, %d'%(ix,iy))

				# TODO: this is too loose. pbase should be STRICTLY contained in proxy
				# check if pillar is on any floor proxy
				pbase_min = np.array([ix,iy,ixmin[2]-0.5]) * unit
				pbase_max = pbase_min+unit
				pbase = aabb_mesh(pbase_min, pbase_max)

				ok = False
				for m in self.scene['proxy_floor']:
					for tri in Mesh.unpack_headerless(m).flatten():
						if thicken_triangle(tri).intersect_convex(pbase):
							ok = True
							break
					if ok:
						break

				if not ok:
					print('rejected as non-floor')
					continue
				
				for iz in range(ixmin[2]+1,ixmin[2]+int(2.5/unit)):
					print('testing',iz)
					i0 = np.array([ix,iy,iz])
					voxel = aabb_mesh(i0*unit, (i0+1)*unit)

					isect = False
					for (o_aabb, o_tris) in aabbs_scene:
						if voxel.intersect_convex(aabb_mesh(*o_aabb)):
							if intersect_aabb_tris(i0*unit, (i0+1)*unit, o_tris):
								isect = True
								break
					
					if isect:
						print('hit something, aborting pillar')
						break
					else:
						voxels[ix-ixmin[0],iy-ixmin[1],iz-ixmin[2]] = True

		# show boundary surface
		def safe_get(p):
			ix,iy,iz = p
			try:
				return voxels[ix,iy,iz]
			except IndexError:
				return False

		print('volume: %.1fm³'%(voxels.sum()*unit**3))

		for ix in range(voxels.shape[0]):
			for iy in range(voxels.shape[1]):
				for iz in range(voxels.shape[2]):
					p = np.array([ix,iy,iz])
					center = (p+ixmin+0.5)*unit

					for dp in np.vstack([np.identity(3),-np.identity(3)]):
						center_face = center+dp*0.5*unit
						scale = (1-0.99*np.abs(dp))*unit

						if safe_get(p) ^ safe_get(p+dp):
							panel = self.mesh_to_actor(aabb_mesh(center_face-scale*0.5, center_face+scale*0.5))
							panel.GetProperty().SetColor(0.5,0.5,1)
							self.renderer.AddActor(panel)



def intersect_aabb_tris(vmin, vmax, tris):
	sep = np.zeros([len(tris)], bool)

	vv = np.array([vmin,vmax])
	vs = np.array([[vv[ix][0],vv[iy][1],vv[iz][2]] for ix in range(2) for iy in range(2) for iz in range(2)])

	# axes
	for n in np.identity(3):
		rs_aabb = np.dot(vs, n)
		rs_tri = (tris*n[np.newaxis,np.newaxis,:]).sum(axis=2)
		sep |= (rs_tri.max(axis=1) < rs_aabb.min()) | (rs_aabb.max() < rs_tri.min(axis=1))

	# normals
	normals = np.cross(tris[:,1]-tris[:,0], tris[:,2]-tris[:,0])
	edges01 = np.cross(tris[:,1]-tris[:,0], normals)
	edges12 = np.cross(tris[:,2]-tris[:,1], normals)
	edges20 = np.cross(tris[:,0]-tris[:,2], normals)

	for ns in [normals, edges01, edges12, edges20]:
		rs_aabb = np.einsum('ij,kj->ki', vs, ns)
		rs_tri = (tris*ns[:,np.newaxis,:]).sum(axis=2)
		sep |= (rs_tri.max(axis=1) < rs_aabb.min(axis=1)) | (rs_aabb.max(axis=1) < rs_tri.min(axis=1))

	return not sep.all()

					


prim_cube = Mesh.load('./prim/cube.mp')
def aabb_mesh(vmin, vmax):
	return prim_cube.apply_cont(lambda vs:vs*(vmax-vmin)+(vmax+vmin)/2)

def thicken_triangle(vs, th=1e-6):
	n = np.cross(vs[1]-vs[0], vs[2]-vs[0])
	d_half = n * (la.norm(n)*th/2)

	vsp = vs+d_half
	vsm = vs-d_half
	vs = np.vstack([vsp,vsm])

	fs = [
		[0,1,2],[5,4,3], # caps
		[0,3,4],[0,4,1], # 0-1 edge
		[1,4,5],[1,5,2], # 1-2 edge
		[2,5,3],[2,3,0]	 # 2-0 edge
	]
	
	return Mesh(vs,fs)


if __name__ == '__main__':
	# add vtkTransform IO to SE3
	def as_vtk(self):
		tr = vtkTransform()
		tr.SetMatrix(self.to_matrix().flatten())
		return tr

	SE3.as_vtk = types.MethodType(as_vtk, None, SE3)

	# normal code
	World(sys.argv[1]).run()
