#!/bin/env python
# coding=utf-8
from __future__ import print_function, division
import cairo
import numpy as np
import scipy.linalg as la
import math
import sys
from vtk import *
import vtk.util.numpy_support as vn
import numpy as np
import time
import types
import random
from tools3d.mesh import Mesh
from tools3d.affine import *
import Queue
import multiprocessing
import imp
from spacetime import parallel
import traceback
import colorsys

#
class Label(object):
	def __init__(self, s):
		self.w = 400
		self.h = 100
		self.data = np.empty([self.h, self.w, 4], np.uint8)
		self.surf = cairo.ImageSurface.create_for_data(self.data, cairo.FORMAT_ARGB32, self.w, self.h)

		ctx = cairo.Context(self.surf)
		ctx.set_source_rgb(0,0,0)
		ctx.paint()
		ctx.set_source_rgb(1,1,1)
		ctx.select_font_face('Monospace')
		ctx.scale(1.5,1.5)
		ctx.translate(10,10)
		ctx.text_path(s)
		ctx.fill()
		self.surf.flush()

class Console(object):
	def __init__(self):
		self.w = 400
		self.h = 300
		self.data = np.empty([self.h, self.w, 4], np.uint8)
		self.surf = cairo.ImageSurface.create_for_data(self.data, cairo.FORMAT_ARGB32, self.w, self.h)

		ctx = cairo.Context(self.surf)
		ctx.set_source_rgb(0,0,0)
		ctx.paint()
		ctx.set_source_rgb(1,1,1)
		ctx.select_font_face('Monospace')
		ctx.scale(1.5,1.5)
		ctx.translate(10,10)
		ctx.text_path(u'こんにちは世界')
		ctx.fill()
		self.surf.flush()
		


class World(object):
	"""
	A VTK window that receives and shows puts
	"""
	def __init__(self, queue):
		self.queue = queue
		self.current = None
		self.odict = {}

		#a renderer and render window
		self.renderer = vtkRenderer()
		self.window = vtkRenderWindow()
		self.window.AddRenderer(self.renderer)
		self.renderer.SetBackground(1,1,1) # white background

		ceiling_light = vtkLight()
		ceiling_light.SetPosition(0, 0, 1)
		ceiling_light.SetAmbientColor(0.3,0.3,0.3)
		self.renderer.AddLight(ceiling_light)

		# add interactor
		self.interactor = vtkRenderWindowInteractor() 
		style = vtkInteractorStyleTrackballCamera()
		self.interactor.SetInteractorStyle(style)
		self.interactor.SetRenderWindow(self.window)

		cam = self.renderer.GetActiveCamera()
		cam.SetPosition(np.array([10,10,1]))
		cam.SetFocalPoint(np.array([0,0,0.9]))
		cam.SetViewUp(np.array([0,0,1]))
		cam.SetViewAngle(90) # vertical fov
		cam.SetClippingRange(0.1,100)
		self.window.Render()

		self.put_floor()
		self.put_model()

		# attach view controller

		# put this before CreateRepeatingTimer otherwise it segfaults
		# http://1045678.n5.nabble.com/iren-gt-CreateRepeatingTimer-segfaults-td1229012.html
		self.interactor.Initialize()

		self.interactor.AddObserver('TimerEvent', self.on_timer)
		self.interactor.CreateRepeatingTimer(30)

	def add_sphere(self,i,c,p):
		src = vtkSphereSource()
		src.SetRadius(0.01)
		src.SetPhiResolution(10)
		src.SetThetaResolution(10)

		m = vtkPolyDataMapper()
		m.SetInputConnection(src.GetOutputPort())

		actor = vtkActor()
		actor.SetMapper(m)
		actor.GetProperty().SetColor(*c)
		actor.GetProperty().SetAmbient(0.1)

		self.renderer.AddActor(actor)

		src.SetCenter(*p)
		self.odict[i] = (actor, src)

	def add_label(self,i,c,p):
		console = Label(p['str'])

		img = vtkImageImport()
		img.SetDataScalarTypeToUnsignedChar()
		img.SetNumberOfScalarComponents(4)
		img.CopyImportVoidPointer(console.data[:,::-1].flatten(), console.w*console.h*4)
		img.SetDataExtent(0,console.w-1,0, console.h-1,0,0)
		img.SetWholeExtent(0,console.w-1,0, console.h-1,0,0)

		plane = vtkPlaneSource()
		plane.SetCenter(0,0,0)
		plane.SetNormal(0,0,1)

		tplane = vtkTextureMapToPlane()
		tplane.SetInputConnection(plane.GetOutputPort())

		tex = vtkTexture()
		tex.SetInputConnection(img.GetOutputPort()) # reader.GetOutputPort())

		m = vtkPolyDataMapper()
		m.SetInputConnection(tplane.GetOutputPort())

		actor = vtkActor()
		actor.SetMapper(m)
		actor.SetTexture(tex)

		self.renderer.AddActor(actor)

		plane.SetCenter(*p['pos'])
		self.odict[i] = (actor, src)

	def on_timer(self, obj, ev):
		if self.current != None:
			radd,rchg,rdel = self.current.now()

			for (i,t,c,p) in radd:
				if t == 'sphere':
					self.add_sphere(i,c,p)
				elif t=='label':
					self.add_label(i,c,p)

			for (i,p) in rchg:
				self.odict[i][1].SetCenter(*p)

			
			if self.current.ended():
				#self.current.restart()
				for o in self.odict.values():
					self.renderer.RemoveActor(o[0])
				self.odict = {}
				self.current = None

			self.window.Render()
		else:
			try:
				self.current = self.queue.get_nowait()
				self.odict = {}
			except Queue.Empty:
				pass

	def put_floor(self):
		"""
		Add 100m by 100m floor grid
		"""
		ix_pt = 0
		pts = vtkPoints()
		lines = vtkCellArray()
		colors = vtk.vtkUnsignedCharArray()
		colors.SetNumberOfComponents(3)
		colors.SetName("colors")

		for i in range(-50,51):
			v0 = np.array([i,-50,0])
			v1 = np.array([i, 50,0])
			pts.InsertNextPoint(v0)
			pts.InsertNextPoint(v1)
			colors.InsertNextTuple3(200,200,200)
			colors.InsertNextTuple3(200,200,200)

			line = vtkLine()
			line.GetPointIds().SetId(0,ix_pt+0)
			line.GetPointIds().SetId(1,ix_pt+1)
			lines.InsertNextCell(line)

			ix_pt += 2

		for j in range(-50,51):
			v0 = np.array([-50,j,0])
			v1 = np.array([ 50,j,0])
			pts.InsertNextPoint(v0)
			pts.InsertNextPoint(v1)
			colors.InsertNextTuple3(200,200,200)
			colors.InsertNextTuple3(200,200,200)

			line = vtkLine()
			line.GetPointIds().SetId(0,ix_pt+0)
			line.GetPointIds().SetId(1,ix_pt+1)
			lines.InsertNextCell(line)

			ix_pt += 2

		# create polydata
		wireframe = vtkPolyData()
		wireframe.SetPoints(pts)
		wireframe.SetLines(lines)
		wireframe.GetPointData().SetScalars(colors)

		# visualize
		mapper = vtkPolyDataMapper()
		mapper.SetInput(wireframe)

		actor = vtkActor()
		actor.SetMapper(mapper)

		self.renderer.AddActor(actor)

	def put_model(self):
		console = Console()

		img = vtkImageImport()
		img.SetDataScalarTypeToUnsignedChar()
		img.SetNumberOfScalarComponents(4)
		img.CopyImportVoidPointer(console.data[:,::-1].flatten(), console.w*console.h*4)
		img.SetDataExtent(0,console.w-1,0, console.h-1,0,0)
		img.SetWholeExtent(0,console.w-1,0, console.h-1,0,0)

		plane = vtkPlaneSource()
		plane.SetCenter(0,0,0)
		plane.SetNormal(0,0,1)

		tplane = vtkTextureMapToPlane()
		tplane.SetInputConnection(plane.GetOutputPort())

		tex = vtkTexture()
		tex.SetInputConnection(img.GetOutputPort()) # reader.GetOutputPort())

		m = vtkPolyDataMapper()
		m.SetInputConnection(tplane.GetOutputPort())

		actor = vtkActor()
		actor.SetMapper(m)
		actor.SetTexture(tex)

		self.renderer.AddActor(actor)

	def run(self):
		self.renderer.ResetCamera()
		self.window.Render()
		self.interactor.Start()
	




class Scenario(object):
	""" Safely picklable motion description. """
	def __init__(self, actors):
		self.actors = actors
		self.t_end = 1 + (max(tb for (tb,_,_,_) in self.actors) if len(self.actors)>0 else 0)
		self.restart()

	def restart(self):
		self.t0 = time.time()
		self.showing = set()

	def now(self):
		t = time.time()-self.t0

		r_add = []
		r_chg = []
		r_del = []

		for (ix,(tbase, tag, c, p)) in enumerate(self.actors):
			if t>tbase:
				if ix in self.showing:
					r_chg.append((ix,p))
				else:
					r_add.append((ix,tag,c,p))
					self.showing.add(ix)

		return (r_add,r_chg,r_del)

	def ended(self):
		t = time.time()-self.t0
		if t>self.t_end:
			return True
		else:
			return False

class HueRange(object):
	def __init__(self, center, half):
		self.center = center
		self.half = half

	@staticmethod
	def whole(center=0):
		return HueRange(center, 0.5)

	def split_n(self, n):
		part = self.half/n
		a0 = self.center-self.half
		return [HueRange(a0+(2*i+1)*part, part) for i in range(n)]

	def center_color(self):
		return colorsys.hsv_to_rgb(self.center%1.0, 0.8, 0.9)


def puts(q, *objs):
	""" Everything is converted to Scenario, and then send via multiprocessing """
	actors = []
	def decompose(obj, tbase, pbase, crange):
		if type(obj) is set or type(obj) is parallel: # parallel
			if len(obj) == 0:
				return
			for (o,ncrange) in zip(obj, crange.split_n(len(obj))):
				decompose(o, tbase, pbase, ncrange)
		elif type(obj) is list: # sequential (200ms)
			for (i,o) in enumerate(obj):
				decompose(o, tbase+0.2*i, pbase, crange) # 200ms is defined from phi (or beta?) effect. the theory states 200ms+ removes sense of motion
		elif type(obj) is tuple and len(obj) == 3: # 3d point
			actors.append((tbase, 'sphere', crange.center_color(), pbase+obj))
		elif type(obj) is np.ndarray and obj.shape==(3,): # 3d point
			actors.append((tbase, 'sphere', crange.center_color(), pbase+obj))
		elif hasattr(obj, 'solidify'):
			decompose(obj.solidify(), tbase, pbase, crange)
		else:
			actors.append((tbase, 'label', crange.center_color(), {'str':str(obj), 'pos': pbase}))
	decompose(parallel(objs), 0, np.zeros(3), HueRange.whole())
	sc = Scenario(actors)
	q.put(sc)
	return sc.t_end

def init_v3d():
	""" Initialize 3D visualization. Return queue to be used with puts """
	# add vtkTransform IO to SE3
	def as_vtk(self):
		tr = vtkTransform()
		tr.SetMatrix(self.to_matrix().flatten())
		return tr
	SE3.as_vtk = types.MethodType(as_vtk, None, SE3)

	q = multiprocessing.Queue()
	def setup():
		World(q).run()

	multiprocessing.Process(target=setup).start()
	return q

if __name__ == '__main__':
	if len(sys.argv) != 2:
		print('usage: ./interact_any.py <script file path>')
	else:
		q = init_v3d()
		path = sys.argv[1]
		cux = {'h':0}
		def acc(d):
			cux['h'] += d

		while True:
			cux['h'] = 0
			try:
				m = imp.load_source('interact', path)
				m.interact(lambda *objs:acc(puts(q,*objs)))
				time.sleep(cux['h'])
			except KeyboardInterrupt:
				break
			except:
				traceback.print_exc()
				time.sleep(2)
