#!/bin/env python
# coding=utf-8
from __future__ import division, print_function
import cairo
import msgpack
import numpy as np
import scipy.linalg as la
import math
import sys
from vtk import *
import vtk.util.numpy_support as vn
import numpy as np
import pyximport
pyximport.install()
import accel
import time
import types
import random
from tools3d.mesh import Mesh
from tools3d.affine import *

def sample_dir(tr, pr, ar, domega_theta=0, domega_phi=0):
	def spherical_to_rot(theta, phi):
		return SE3.rotate(np.array([0,0,1]), phi) * SE3.rotate(np.array([0,1,0]), theta)

	def add_spherical(theta, phi, dtheta, dphi):
		"""
		Return new (theta,phi) moved by (dtheta, dphi).
		Reference direction for dphi is determined by unit phi vector.
		"""
		return spherical_to_rot(theta, phi) * spherical_to_rot(dtheta, dphi)

	theta = random.random()*(tr[1]-tr[0])+tr[0]
	phi = random.random()*(pr[1]-pr[0])+pr[0]
	alpha = random.random()*(ar[1]-ar[0])+ar[0]

	return add_spherical(theta, phi, domega_theta, alpha + domega_phi) * np.array([0,0,1])

class World(object):
	"""
	A VTK window that contains interactive visualization of solid angle.
	"""
	def __init__(self):
		#a renderer and render window
		self.renderer = vtkRenderer()
		self.window = vtkRenderWindow()
		self.window.AddRenderer(self.renderer)
		self.renderer.SetBackground(1,1,1) # white background

		ceiling_light = vtkLight()
		ceiling_light.SetPosition(0, 0, 1)
		ceiling_light.SetAmbientColor(0.3,0.3,0.3)
		self.renderer.AddLight(ceiling_light)

		# add interactor
		self.interactor = vtkRenderWindowInteractor() 
		style = vtkInteractorStyleTrackballCamera()
		self.interactor.SetInteractorStyle(style)
		self.interactor.SetRenderWindow(self.window)

		# prepare model
		self.put_model()

		# attach view controller

		# put this before CreateRepeatingTimer otherwise it segfaults
		# http://1045678.n5.nabble.com/iren-gt-CreateRepeatingTimer-segfaults-td1229012.html
		self.interactor.Initialize()



	def on_timer(self, obj, ev):
		self.update_raster()
		self.window.Render()

	def on_keypress(self, obj, ev):
		table = {
			# translation
			'Up': SE3.translate(0.1*np.array([0,1,0])),
			'Down': SE3.translate(0.1*np.array([0,-1,0])),
			'Left': SE3.translate(0.1*np.array([-1,0,0])),
			'Right': SE3.translate(0.1*np.array([1,0,0])),
			'space': SE3.translate(0.1*np.array([0,0,1])),
			'Shift_L': SE3.translate(0.1*np.array([0,0,-1])),
			# rotation
			'Control_L': SE3.rotate(np.array([1,0,0]), math.pi*0.5),
			'Alt_L': SE3.rotate(np.array([0,1,0]), math.pi*0.5),
			'Shift_R': SE3.rotate(np.array([0,0,1]), math.pi*0.5)
		}

		k = obj.GetKeySym()
		if k in table:
			self.cam2world = self.cam2world * table[k]

			self.drw.notify_view(self.cam2world)
			self.update_view()
			self.window.Render()
		else:
			print('unhandled',k)

	def run(self):
		self.renderer.ResetCamera()
		self.window.Render()
		self.interactor.Start()

	def put_model(self):
		ddeg = 10
		for phi0 in [ddeg*ix for ix in range(int(180/ddeg)+1)]:
			phi1 = phi0+0.5
			src = vtkSphereSource()
			src.SetRadius(1.001)
			src.SetStartPhi(phi0)
			src.SetEndPhi(phi1)
			src.SetPhiResolution(5)
			src.SetThetaResolution(50)

			transformFilter = vtkTransformPolyDataFilter()
			transformFilter.SetTransform(vtkTransform())
			transformFilter.SetInputConnection(src.GetOutputPort())

			m = vtkPolyDataMapper()
			m.SetInputConnection(transformFilter.GetOutputPort())

			a = vtkActor()
			a.SetMapper(m)
			a.GetProperty().SetAmbient(0.1)

			self.renderer.AddActor(a)

		for theta0 in [ddeg*ix for ix in range(int(360/ddeg)+1)]:
			theta1 = theta0+0.5
			src = vtkSphereSource()
			src.SetRadius(1.001)
			src.SetStartTheta(theta0)
			src.SetEndTheta(theta1)
			src.SetPhiResolution(50)
			src.SetThetaResolution(5)

			transformFilter = vtkTransformPolyDataFilter()
			transformFilter.SetTransform(vtkTransform())
			transformFilter.SetInputConnection(src.GetOutputPort())

			m = vtkPolyDataMapper()
			m.SetInputConnection(transformFilter.GetOutputPort())

			a = vtkActor()
			a.SetMapper(m)
			a.GetProperty().SetAmbient(0.1)

			self.renderer.AddActor(a)

		# add point cloud
		xyz = []
		rgb = []

		# grid
		itheta = 3
		iphi = 5
		ialpha = 8

		ntheta = 8
		nphi = 16
		nalpha = 16

		# omega
		om_itheta = 5
		om_iphi = 10
		om_ntheta = 16
		om_nphi = 32


		for i in range(20000):
			zeta_theta = np.random.random()
			zeta_phi = np.random.random()

			om_theta = math.pi*(zeta_theta+om_itheta)/om_ntheta
			om_phi = 2*math.pi*(zeta_phi+om_iphi)/om_nphi

			d = sample_dir(
				[math.pi*itheta/ntheta, math.pi*(itheta+1)/ntheta],
				[2*math.pi*iphi/nphi, 2*math.pi*(iphi+1)/nphi],
				[2*math.pi*ialpha/nalpha, 2*math.pi*(ialpha+1)/nalpha],
				om_theta, om_phi
			)

			c = [
				zeta_theta,
				zeta_phi,
				0.5
			]
			xyz.append(d)
			rgb.append(c)

			d = [
				math.sin(om_theta)*math.cos(om_phi),
				math.sin(om_theta)*math.sin(om_phi),
				math.cos(om_theta)]

			xyz.append(d)
			rgb.append(map(lambda x:x*0.5, c))

		xyz = np.array(xyz)
		rgb = np.array(rgb)*255

		pts = vtk.vtkPoints()
		verts = vtk.vtkCellArray()
		colors = vtk.vtkUnsignedCharArray()
		colors.SetNumberOfComponents(3)
		colors.SetName("colors")

		for (c,p) in zip(rgb.reshape([-1,3]), xyz.reshape([-1,3])):
			ix = pts.InsertNextPoint(p)
			verts.InsertNextCell(1)
			verts.InsertCellPoint(ix)
			colors.InsertNextTuple3(*c)
		
		d = vtk.vtkPolyData()
		d.SetPoints(pts)
		d.SetVerts(verts)
		d.GetPointData().SetScalars(colors)

		m = vtk.vtkPolyDataMapper()
		m.SetInput(d)

		a = vtk.vtkActor()
		a.SetMapper(m)

		self.renderer.AddActor(a)
		cloud_actor = a


if __name__ == '__main__':
	# add vtkTransform IO to SE3
	def as_vtk(self):
		tr = vtkTransform()
		tr.SetMatrix(self.to_matrix().flatten())
		return tr

	SE3.as_vtk = types.MethodType(as_vtk, None, SE3)

	# normal code
	World().run()
