#!/bin/env python
# coding=utf-8
"""
Polygon reduction experiment.

references
http://dev.gameres.com/Program/Visual/3D/PolygonReduction.pdf
http://research.microsoft.com/en-us/um/people/hoppe/pm.pdf
"""
from __future__ import division, print_function
import cairo
import msgpack
import numpy as np
import scipy.linalg as la
import math
import sys
from vtk import *
import vtk.util.numpy_support as vn
import numpy as np
import pyximport
pyximport.install()
import accel
import time
import types
import random
from tools3d.mesh import Mesh
from tools3d.affine import *

def collapse_edge(mesh, method='cost-len'):
	# create edge list
	edges = np.empty([3*len(mesh.fs), 2], int)
	edges[0::3] = mesh.fs[:,:2]
	edges[1::3] = mesh.fs[:,1:]
	edges[2::3] = mesh.fs[:,::-2]

	# select edge to collapse
	if method == 'random':
		ixe = random.randrange(len(edges))
	elif method == 'cost-len':
		elen = ((mesh.vs[edges[:,0]]-mesh.vs[edges[:,1]])**2).sum(axis=1)**0.5
		ixe = np.argmin(elen)
	else:
		raise ValueError('unknown edge collapse method '+method)

	s,t = edges[ixe]

	# remove edge
	# count s,t in each face (2:removal 1:replace 0:normal)
	vcount = ((mesh.fs == s) | (mesh.fs == t)).sum(axis=1)
	fs_new = mesh.fs[vcount!=2]

	# replace s -> t
	fs_new = fs_new * (fs_new != s) + t * (fs_new == s)

	return (Mesh(mesh.vs, fs_new), mesh.vs[t])


def pack_vertices(mesh):
	"""
	Remove unused vertices.
	"""
	return mesh

class World(object):
	"""
	A VTK window that contains interactive visualization of solid angle.
	"""
	def __init__(self, path):
		self.scene_path = path
		self.scene = msgpack.unpack(open(path,'rb'))
		
		#a renderer and render window
		self.renderer = vtkRenderer()
		self.window = vtkRenderWindow()
		self.window.AddRenderer(self.renderer)
		self.renderer.SetBackground(.1,.1,.1) # dark gray background

		ceiling_light = vtkLight()
		ceiling_light.SetPosition(0, 0, 100)
		ceiling_light.SetAmbientColor(0.3,0.3,0.3)
		self.renderer.AddLight(ceiling_light)

		# add interactor
		self.interactor = vtkRenderWindowInteractor() 
		style = vtkInteractorStyleTrackballCamera()
		self.interactor.SetInteractorStyle(style)
		self.interactor.SetRenderWindow(self.window)

		for name in ['KeyPressEvent']:
			self.interactor.RemoveObservers(name)

		self.view_move = None
		self.verts = []
		self.interactor.AddObserver('KeyPressEvent', self.on_key_press)

		# prepare model
		sel = 'mesh30'
		for obj in self.scene['objects']:
			name = obj['name']
			mesh = Mesh.unpack_headerless(obj['mesh'])
			print(name, len(mesh))

			if name == sel:
				actor = self.mesh_to_actor(mesh, True)
				self.renderer.AddActor(actor)

				self.current_m = mesh
				self.current_a = actor

		src = vtkSphereSource()
		src.SetRadius(0.03)
		src.SetPhiResolution(10)
		src.SetThetaResolution(10)
		 
		m = vtkPolyDataMapper()
		m.SetInputConnection(src.GetOutputPort())

		actor = vtkActor()
		actor.SetMapper(m)
		actor.GetProperty().SetColor(1,0.5,0.5)
		actor.GetProperty().SetAmbient(0.1)

		self.renderer.AddActor(actor)
		self.cursor = src

		# put this before CreateRepeatingTimer otherwise it segfaults
		# http://1045678.n5.nabble.com/iren-gt-CreateRepeatingTimer-segfaults-td1229012.html
		self.interactor.Initialize()

		self.interactor.AddObserver('TimerEvent', self.on_timer)
		self.interactor.CreateRepeatingTimer(100)


	def on_timer(self, obj, ev):
		pass

	def on_key_press(self, obj, ev):
		table_avatar = {
			# translation
			'Up': SE3.translate(0.1*np.array([0,1,0])),
			'Down': SE3.translate(0.1*np.array([0,-1,0])),
			'Left': SE3.translate(0.1*np.array([-1,0,0])),
			'Right': SE3.translate(0.1*np.array([1,0,0])),
			'space': SE3.translate(0.1*np.array([0,0,1])),
			'Shift_L': SE3.translate(0.1*np.array([0,0,-1]))
		}

		k = obj.GetKeySym()
		if k == 'z': # remove 1 edge
			self.current_m, pt = collapse_edge(self.current_m)
			self.renderer.RemoveActor(self.current_a)

			self.current_a = self.mesh_to_actor(self.current_m)
			self.renderer.AddActor(self.current_a)

			self.cursor.SetCenter(pt)
			self.window.Render()

			print('#v=%d #f=%d'%(len(self.current_m.vs), len(self.current_m.fs)))
		elif k == 'x': # remove until face count become 1000
			while len(self.current_m.fs)>1000:
				self.current_m, pt = collapse_edge(self.current_m)
				print('#v=%d #f=%d'%(len(self.current_m.vs), len(self.current_m.fs)))

			self.renderer.RemoveActor(self.current_a)

			self.current_a = self.mesh_to_actor(self.current_m)
			self.renderer.AddActor(self.current_a)

			self.cursor.SetCenter(pt)
			self.window.Render()
		elif k == 'q':
			import sys
			sys.exit(0)
		else:
			print('unhandled',k)

	def run(self):
		self.renderer.ResetCamera()
		self.window.Render()
		self.interactor.Start()

	def mesh_to_actor(self, mesh, as_wireframe=True):
		"""
		Convert given mesh to an vtkActor.

		cf. Fast way to convert numpy array to vtk objects:
		http://vtk.1045678.n5.nabble.com/numpy-to-vtk-and-vtkPolyData-td4949286.html
		"""
		poly = vtkPolyData()

		# points
		pts = vtkPoints()
		pts.SetData(vn.numpy_to_vtk(mesh.vs))
		poly.SetPoints(pts)

		# map each face to 3 lines
		if as_wireframe:
			lines_raw = np.empty([len(mesh.fs)*3, 3], np.int64)
			lines_raw[:,0] = 2
			lines_raw[0::3,1:] = mesh.fs[:,:2]
			lines_raw[1::3,1:] = mesh.fs[:,1:]
			lines_raw[2::3,1:] = mesh.fs[:,::-2]

			lines = vtkCellArray()
			lines.SetCells(len(lines_raw), vn.numpy_to_vtkIdTypeArray(lines_raw))
			poly.SetLines(lines)

			poly.dont_collect = [mesh.vs, lines_raw]
		else:
			tris_raw = np.empty([len(mesh.fs), 4], np.int64)
			tris_raw[:,0] = 3
			tris_raw[:,1:] = mesh.fs

			tris = vtkCellArray()
			tris.SetCells(len(tris_raw), vn.numpy_to_vtkIdTypeArray(tris_raw))
			poly.SetPolys(tris)

			poly.dont_collect = [mesh.vs, tris_raw]

		# wrap as an Actor
		mapper = vtkPolyDataMapper()
		mapper.SetInput(poly)

		actor = vtkActor()
		actor.SetMapper(mapper)

		return actor



if __name__ == '__main__':
	# add vtkTransform IO to SE3
	def as_vtk(self):
		tr = vtkTransform()
		tr.SetMatrix(self.to_matrix().flatten())
		return tr

	SE3.as_vtk = types.MethodType(as_vtk, None, SE3)

	# normal code
	World(sys.argv[1]).run()
