#!/bin/env python
# coding=utf-8
from __future__ import print_function, division
import sys
import msgpack
import math
import random
import numpy as np
import scipy.linalg as la
import scipy.misc
from tools3d.affine import *
from tools3d.mesh import *
from scene import Scene
from index import *
import operator
import os.path
import traceback
from fvxl import dumpSimpleFVXL, loadSimpleFVXL

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.font_manager as fm

def prod(xs):
	return reduce(operator.mul, xs, 1)

def multi_any_el(ls, axes):
	"""
	Project ls to given axes by logical OR.
	retval == ls when axes=(0,1,2,3,4,5)
	"""
	def f(x):
		return tuple(x[axis] for axis in axes)
	return set(map(f, ls))


def multi_any(array, axes):
	red = array
	for i in reversed(list(axes)):
		red = red.any(axis=i)
	return red

def multi_sum(array, axes):
	red = array
	for i in reversed(list(axes)):
		red = red.sum(axis=i)
	return red


class DeviceModel(object):
	""" This class is deprecated. """
	px_per_dx = 100
	def __init__(self, fovx=math.radians(120), fovy=math.radians(90), error=LinearErrorModel(0.2/4)):
		self.have_accel = True
		self.fovx = fovx
		self.fovy = fovy
		self.error = error
		self.ns_map = self.gen_ns_map()

	def get_field(self, aunit):
		return DiscreteCameraField(self.fovx, self.fovy, aunit)

	def get_desc(self):
		return 'eps=%s / FOV:%d * %d'%(self.error.get_desc(), int(math.degrees(self.fovx)), int(math.degrees(self.fovy)))

	def gen_ns_map(self):
		"""
		Return (height,width,3) normal vectors in local coordinates.
		These normals denote directions of each pixel.
		"""
		# ns = np.array([[]])
		pass
		
	def simulate_dense(self, acc, config):

		sx = math.tan(self.fovx/2)*2
		sy = math.tan(self.fovy/2)*2

		w = int(sx * px_per_dx)
		h = int(sy * px_per_dx)
		result = np.zeros([h,w],float)

		ray_f = config * np.zeros(3)
		config_rot = config.extract_rotation()
		for iy in range(h):
			for ix in range(w):
				dx = sx*(ix/w-0.5)
				dy = sy*(iy/h-0.5)
				d = np.array([dx,dy,1])
				d /= la.norm(d)
				result[iy,ix] = acc.raycast(ray_f, config_rot * d)

		return result

	def rotate(self, direction, dmap, orientation, iori):
		""" "Safely" rotate dmap with orientation(iori) and return it as discretized with direction. """
		r_pixel = 0.01
		rot, r_rot = orientation.cone(iori)
		r = r_pixel + r_rot

		# ns = np.dot(dmap, dmap) -> discretize and expand with r


class ODeviceModel(object):
	def __init__(self, error=LinearErrorModel(0)): # 0.2/4
		self.have_accel = True
		self.error = error
		self.px_per_rad = 30 # 15
		
		# derived attributes
		self.width = int(2 * math.pi * self.px_per_rad)
		self.height = int(math.pi * self.px_per_rad)

		self.map_ns = self.gen_ns()

	def gen_ns(self):
		map_ns = np.empty([self.height, self.width, 3], np.float32)

		for it in xrange(self.height):
			for ip in xrange(self.width):
				theta = math.pi * it/self.height
				phi = 2 * math.pi * ip/self.width

				map_ns[it,ip] = np.array([
					math.sin(theta)*math.cos(phi),
					math.sin(theta)*math.sin(phi),
					math.cos(theta)
					])

		return map_ns

	def get_field(self, aunit):
		return DiscreteDirection(aunit)

	def get_desc(self):
		return 'eps=%s / FOV: omnidirectional'%self.error.get_desc()
		
	def simulate_dense(self, acc, pose):
		result = np.zeros([self.height,self.width],np.float32)

		ray_f = pose * np.zeros(3)
		pose_rot = pose.extract_rotation()
		for it in xrange(self.height):
			for ip in xrange(self.width):
				result[it,ip] = acc.raycast(ray_f, pose_rot * self.map_ns[it,ip])

		# limit to 120 degree
		#result[(self.height*2)//3:: , 0::2] = 0
		#result[(self.height*2)//3:: , 1::2] = 1e3

		return result

	def rotate_discretize(self, direction, dmap, orientation, iori):
		"""
		Rotate dmap with orientation(iori), and discretize it with direction.
		All operations are consavatively calculated.

		Returned shape is (len(direction),2), which contains (min,max) for each cell.
		"""
		r_pixel = math.sqrt(2)*(1/self.px_per_rad)/2
		rot, r_rot = orientation.cone(iori)

		ndirs, r_dir = direction.cones()
		cos_radius = math.cos(r_pixel + r_rot + r_dir)

		ndirs_local = np.einsum('ij,di->dj', rot, ndirs) # la.inv(rot)??? TODO
		targets = np.einsum('yxi,di->dyx', self.map_ns, ndirs_local) >= cos_radius

		result = np.empty([len(direction),2], np.float32)
		for idir in xrange(len(direction)):
			vs = dmap[targets[idir]]

			if vs.size>0:
				result[idir,0] = vs.min()
				result[idir,1] = vs.max()
			else:
				result[idir,0] = 0
				result[idir,1] = 1e6

		return result

def simulate_localize_session(index, path, dev, path_ixs, base_p='', fixed_ori_set=None):
	valid_count = index.count_valid_cells()
	print('counted valid', valid_count)

	scene = Scene.unpack_headerless(index.o['scene'])
	acc = scene.get_accel()

	xds = [] # localized region over time for X,Y,Z
	durs = []
	bits = []
	for (ipose,pose) in enumerate(path):
		dmap = dev.simulate_dense(acc, pose)
		if ipose % 25 == 0:
			scipy.misc.imsave(base_p+'pose%03d-render.png'%ipose, (dmap/15*255).astype(np.uint8))

		t0 = time.time()
	   
		if fixed_ori_set == None:
			ori_filter = possible_orientations(index.orientation, pose.extract_rotation().I * np.array([0,0,-1]))
		else:
			ori_filter = fixed_ori_set
		
		locs = index.localize(dev, dmap, ori_filter)
		xds.append(locs)
		t1 = time.time()
		durs.append(t1-t0)

		# stat(new)
		locs_el = index.to_elements(locs)
		n_original = len(locs_el)
		n_factored6 = prod(len(multi_any_el(locs_el, [i])) for i in range(6))
		n_factored2 = len(multi_any_el(locs_el, [0,1,2])) * len(multi_any_el(locs_el, [3,4,5]))
		
		try:
			e = entropy(index, locs)
		except ValueError:
			e = -1
			print('impossible happened: check your config / scene data / parameters etc.')
		bits.append(e)
		try:
			print(pose,
				'#gravity/#ori=', len(ori_filter)/len(index.orientation) if ori_filter else 1,
				'orig/valid=', n_original/valid_count,
				'orig/factored(6)=', n_original/n_factored6,
				'orig/factored(2)=', n_original/n_factored2,
				int(1e3*(t1-t0)),'ms')
			print(e,'bit')
		except ZeroDivisionError:
			print('impossible happened: check your config / scene data / parameters etc.')

	return (xds, durs, bits)

def possible_orientations(orientation, down_local):
	""" Return possible orientation indices that can transform given local gravity vector to global gravity vector (0,0,-1). """
	res = set()
	for iori in xrange(len(orientation)):
		trans,radius = orientation.cone(iori)
		down_global = np.dot(trans, down_local)
		cos = -down_global[2]
		if cos >= math.cos(radius):
			res.add(iori)
	return res



def vect_to_trans(vect):
	return (
		SE3.translate(vect[:3]) *
		SE3.rotate(np.array([0,0,1]), vect[3+1]) *
		SE3.rotate(np.array([0,1,0]), vect[3+0]) *
		SE3.rotate(np.array([0,0,1]), vect[3+2]))


def entropy(index, result):
	""" Return KL-divergence based entropy of the result. """
	flags = index.to_legacy(result)
	theta = multi_sum(flags, [0,1,2,4,5])

	n = index.count_valid_cells()
	m = theta.sum()
	naive_entropy = -math.log(m/n,2)
	print('n,m,naive',n,m,naive_entropy)

	ntheta = index.orientation.ntheta
	dtheta = math.pi/ntheta

	weight = (ntheta/n)*math.sin(dtheta*0.5)*np.sin((np.array(range(ntheta), float)+0.5)*dtheta)

	return -math.log((weight*theta).sum(), 2)


def generate_random(index, n_frames=100):
	""" Return (path_ixs, path_vects). """
	# index*scale + offset = value
	# 0<=index<shape
	offset = np.zeros([6])
	scale = np.zeros([6])
	offset[:3] = np.array(index.grid['origin'])	 #+ [0,0,0.225] # TODO remove
	scale[:3] = index.grid['lunit']
	scale[3] = math.pi/index.orientation.ntheta
	scale[4] = 2*math.pi/index.orientation.nphi
	scale[5] = 2*math.pi/index.orientation.nalpha
	shape = index.shape

	#
	def sample_pos():
		""" Uniformly sample valid position tuple. """
		ixs = tuple(np.floor(np.random.rand(3)*index.pos_shape).astype(int))
		if index.samples_full[ixs[0], ixs[1], ixs[2], 0, 0, 0] > 1e3:
			return sample_pos()
		else:
			return ixs

	def sample_rot():
		return tuple(np.floor(np.random.rand(3)*index.rot_shape).astype(int))

	path_ixs = []
	for i in xrange(n_frames):
		p = sample_pos()
		r = sample_rot()
		path_ixs.append(p+r)


	path_vects = (np.array(path_ixs)+np.random.rand(n_frames, 6)) * scale + offset
	

	return (path_ixs, path_vects)



def evaluate_localization(path_index, path_figure):
	n_frames = 100
	#n_frames = 1

	# random sampling to generate poses
	print('== Preprocess')
	index = Index.load_from(path_index)

	# index*scale + offset = value
	# 0<=index<shape
	offset = np.zeros([6])
	scale = np.zeros([6])
	offset[:3] = np.array(index.grid['origin'])	 #+ [0,0,0.225] # TODO remove
	scale[:3] = index.grid['lunit']
	scale[3] = math.pi/index.orientation.ntheta
	scale[4] = 2*math.pi/index.orientation.nphi
	scale[5] = 2*math.pi/index.orientation.nalpha
	shape = index.shape


	path_ixs, path_vects = generate_random(index, n_frames)
	path_poses = map(vect_to_trans, path_vects)


	dev = ODeviceModel()

	print('== Simulation')
	n_dist_valid = index.count_valid_cells()
	dists, durs, bits = simulate_localize_session(index, path_poses, dev, path_ixs, path_figure)
	print(path_ixs)
	print(path_vects)
	print(durs)
	print(bits)

	fs_contained = []
	for (ixs,dist) in zip(path_ixs, dists):
		try:
			ipos = np.ravel_multi_index(ixs[:3], index.pos_shape)
			iori = np.ravel_multi_index(ixs[3:], index.orientation.shape)
			fs_contained.append((ipos,iori) in dist)
		except ValueError:
			fs_contained.append(False)


	print('== Visualization')
	cm_to_inch = 0.393700787

	print('=== XY')
	for i in range(n_frames):
		plt.figure(figsize=(8*cm_to_inch, 5*cm_to_inch))
		tx = path_vects[i][0]
		ty = path_vects[i][1]
		tz = path_vects[i][2]

		xy = multi_sum(index.to_legacy(dists[i]), {2,3,4,5})
		xr = (offset[0], offset[0]+scale[0]*shape[0])
		yr = (offset[1], offset[1]+scale[1]*shape[1])

		plt.title('(x,y,z)=(%.1f,%.1f,%.1f)'%(tx, ty, tz))
		plt.xlabel('y [m]')
		plt.ylabel('x [m]')

		plt.xlim(*yr)
		plt.ylim(*xr)
		plt.imshow(xy[::-1],extent=(yr+xr),interpolation='nearest',cmap=plt.get_cmap('hot'))
		plt.gca().invert_yaxis()
		#plt.grid(True)
		plt.colorbar()
		plt.scatter([ty],[tx],s=15,facecolor='white')

		pf = 'slices-%.1fbit-%s'%(bits[i], path_figure)
		plt.tight_layout()
		plt.savefig(pf, dpi=600)
		print('{{%s}}'%pf)



if __name__ == '__main__':
	evaluate_localization(*sys.argv[1:])
