#!/usr/bin/python
from __future__ import division, print_function
import numpy as np
import math
import msgpack
import sys
import traceback
import tools3d.mesh as mesh
import re


def remove_comment(s):
	ix = s.find('#')
	if ix<0:
		return s
	else:
		return s[:ix]

def parse_obj_naive(file_in):
	"""
	Return ([object],vertices,faces). No tesselation or index check is performed.
	Index is converted from 1-origin to 0-origin in returned faces.

	group:
		v*
		vt*
		g
		usemtl
		f*
	"""
	objects = []
	v_list = []
	f_list = []

	commands = ['v','vt','g','f','usemtl','mtllib','o','s']
	for line in file_in:
		ws = remove_comment(line).split()
		if not ws:
			continue

		command = ws[0]
		args = ws[1:]

		#if command != prev_command:
		if command == 'g' or command == 'o':
			if objects:
				objects[-1]['f_end'] = len(f_list)

			objects.append({
				'name': args[0],
				'f_start': len(f_list)
				})
			ix_next = len(v_list)
		elif command == 'v':
			v_list.append(map(float,args))
		elif command == 'f':
			f_list.append(map(lambda w:int(w.split('/')[0])-1, args))
		elif command not in commands:
			raise NotImplementedError('unknown command %s'%command)

	if objects:
		objects[-1]['f_end'] = len(f_list)
	return (objects, v_list, f_list)

def swap_axis(mesh):
	return mesh.apply_cont(lambda vs:np.array([vs[:,2],vs[:,0],vs[:,1]]).T)

def parse(path_in):
	complex_objs, complex_vs, complex_fs = parse_obj_naive(open(path_in,'r'))
	print('#obj=%d #vert=%d #face=%d'%(len(complex_objs), len(complex_vs), len(complex_fs)))

	new_objs = []
	for obj in complex_objs:
		try:
			# tesselate to triangles
			tris = []
			for f in complex_fs[obj['f_start']:obj['f_end']]:
				if len(f) == 3:
					tris.append(f)
				elif len(f) == 4:
					tris.append(f[:3])
					tris.append([f[0],f[2],f[3]])
				else:
					raise NotImplementedError("couldn't tesselate %s"%f)

			# convert vertex index to object-local index
			vix_min = min(vix for t in tris for vix in t)
			vix_max = max(vix for t in tris for vix in t)

			new_obj = {
				'name': obj['name'],
				'mesh': swap_axis(mesh.Mesh(complex_vs[vix_min:vix_max+1], [[vix-vix_min for vix in tri] for tri in tris]))
			}

			new_objs.append(new_obj)
		except:
			print('ignoring object %s since following exception was raised'%obj['name'])
			traceback.print_exc()

	return new_objs

def convert(path_in, path_out):
	objs = parse(path_in)
	objs.sort(key=lambda o:o['name'])

	for obj in objs:
		obj['mesh'] = obj['mesh'].pack_headerless()

	msgpack.pack(objs, open(path_out,'wb'))

def auto_group(objs):
	def to_base(name):
		return re.match('([A-Za-z_]+)[0-9]*', name).groups()[0]

	obj_groups = {}
	for obj in objs:
		obj_groups.setdefault(to_base(obj['name']), []).append(obj)

	for (og_name, og_objs) in obj_groups.items():
		print(og_name, len(og_objs))


if __name__ == '__main__':
	convert(*sys.argv[1:])
