# coding=utf-8
from __future__ import print_function, division
import sys
import msgpack
import math
import random
import numpy as np
import scipy.linalg as la
from tools3d.affine import *
from tools3d.mesh import *
import pyximport
pyximport.install()
import accel
import os.path


class Scene(object):
	"""
	Annotated Scene
	"""
	def __init__(self, o):
		self.original = o
		self.objects = []

		for obj in o['objects']:
			m = Mesh.unpack_headerless(obj['mesh'])
			aabb = (m.vs.min(axis=0), m.vs.max(axis=0))
			self.objects.append({
				'mesh': m,
				'aabb': aabb
				})

	@staticmethod
	def load(path):
		return Scene.unpack_headerless(msgpack.unpack(open(path,'rb')))

	@staticmethod
	def unpack_headerless(o):
		return Scene(o)

	def pack_headerless(self):
		return self.original

	def gen_region(self, unit=0.5):
		""" Use scene to generate empty voxels.

		For any index in an axis, that voxel spans [index*unit, (index+1)*unit).

		Generated voxel is
		1. strictrly contained in floor proxy
		2. does not intersect with mesh
		"""
		scene = self.original

		access_height = 2.5 # accesible height from floor plane

		vs = np.vstack([Mesh.unpack_headerless(m).vs for m in scene['proxy_floor']])
		aabb_min, aabb_max = vs.min(axis=0), vs.max(axis=0)

		imin = np.floor(aabb_min/unit).astype(np.int)
		imax = np.ceil((aabb_max+[0,0,access_height])/unit).astype(np.int)

		# flatten mesh of each object
		ttris_scene = [(obj['aabb'],obj['mesh'].flatten()) for obj in self.objects]

		# create pillars
		voxels = np.zeros(imax-imin, bool)
		for ix in xrange(voxels.shape[0]):
			for iy in xrange(voxels.shape[1]):
				# TODO: this is too loose. pbase should be STRICTLY containec in proxy
				# check if pillar is on any floor proxy
				pbase_min = (imin + [ix,iy,-0.5]) * unit
				pbase_max = pbase_min+unit
				pbase = aabb_mesh(pbase_min, pbase_max)

				ok = False
				for m in scene['proxy_floor']:
					for tri in Mesh.unpack_headerless(m).flatten():
						if thicken_triangle(tri).intersect_convex(pbase):
							ok = True
							break
					if ok:
						break

				if not ok:
					continue
				
				for iz in xrange(1, voxels.shape[2]): # skip lowest layer
					vx_min = (imin+[ix,iy,iz]) * unit
					vx_max = vx_min + unit
					voxel = aabb_mesh(vx_min, vx_max)

					isect = False
					for (o_aabb, o_tris) in ttris_scene:
						if intersect_aabb_aabb(vx_min, vx_max, *o_aabb):
							if intersect_aabb_tris(vx_min, vx_max, o_tris):
								isect = True
								break
					
					if isect:
						break
					else:
						voxels[ix,iy,iz] = True

		print('volume: %.1fm³'%(voxels.sum()*unit**3))
		return self.compact_voxels(imin, voxels)

	def compact_voxels(self, ixmin, voxels):
		rs = []
		for axis in range(3):
			vx_reduced = voxels
			for axis_reduce in reversed(sorted(list({0,1,2} - {axis}))):
				vx_reduced = vx_reduced.any(axis=axis_reduce)

			ixs_needed = np.arange(len(vx_reduced))[vx_reduced]
			v0 = ixs_needed.min()
			v1 = ixs_needed.max()

			rs.append((v0,v1+1))

		voxels_c = voxels[rs[0][0]:rs[0][1], rs[1][0]:rs[1][1], rs[2][0]:rs[2][1]]
		ixmin_c = ixmin + [rs[0][0], rs[1][0], rs[2][0]]
		
		return (ixmin_c, voxels_c.copy())

	def get_accel(self):
		return SceneAccel.from_scene_objects(self.objects)

class SceneAccel(object):
	def __init__(self, objects):
		self.objects = objects

	@staticmethod
	def from_scene_objects(s_objects):
		objects = []
		for obj in s_objects:
			mesh = obj['mesh']
			vmin, vmax = obj['aabb']

			objects.append((
				accel.BIHAccel(mesh),
				accel.AABBAccel(vmin, vmax),
				(vmin,vmax)
				))

		return SceneAccel(objects)

	def half_space(self, n, offset):
		"""
		Return subset of self that correctly work when ray satisfy:

		dot(n,fr) >= offset
		dot(n,d) >= 0
		"""
		def may_intsect(obj):
			acc, acc_aabb, vs_aabb = obj
			vs = [(vs_aabb[sx][0],vs_aabb[sy][1],vs_aabb[sz][2])
				for sx in range(2) for sy in range(2) for sz in range(2)]
			return (np.dot(vs, n) >= offset).any()
		
		return SceneAccel(filter(may_intsect, self.objects))
	
	def raycast(self, fr, d):
		t = 1e3
		for (acc, acc_aabb, vs_aabb) in self.objects:
			if acc_aabb.raycast(fr,d)<t:
				tc = acc.raycast(fr,d)
				if tc<t:
					t = tc

		return t


prim_cube = Mesh.load(os.path.join(os.path.dirname(__file__),'prim/cube.mp'))
def aabb_mesh(vmin, vmax):
	return prim_cube.apply_cont(lambda vs:vs*(vmax-vmin)+(vmax+vmin)/2)

def intersect_aabb_aabb(vmin, vmax, umin, umax):
	return not ((vmax<umin) | (umax<vmin)).any()


def intersect_aabb_tris(vmin, vmax, tris):
	sep = np.zeros([len(tris)], bool)

	vv = np.array([vmin,vmax])
	vs = np.array([[vv[ix][0],vv[iy][1],vv[iz][2]] for ix in range(2) for iy in range(2) for iz in range(2)])

	# axes
	for n in np.identity(3):
		rs_aabb = np.dot(vs, n)
		rs_tri = (tris*n[np.newaxis,np.newaxis,:]).sum(axis=2)
		sep |= (rs_tri.max(axis=1) < rs_aabb.min()) | (rs_aabb.max() < rs_tri.min(axis=1))

	# normals
	normals = np.cross(tris[:,1]-tris[:,0], tris[:,2]-tris[:,0])
	edges01 = np.cross(tris[:,1]-tris[:,0], normals)
	edges12 = np.cross(tris[:,2]-tris[:,1], normals)
	edges20 = np.cross(tris[:,0]-tris[:,2], normals)

	for ns in [normals, edges01, edges12, edges20]:
		rs_aabb = np.einsum('ij,kj->ki', vs, ns)
		rs_tri = (tris*ns[:,np.newaxis,:]).sum(axis=2)
		sep |= (rs_tri.max(axis=1) < rs_aabb.min(axis=1)) | (rs_aabb.max(axis=1) < rs_tri.min(axis=1))

	return not sep.all()

					

def thicken_triangle(vs, th=1e-6):
	n = np.cross(vs[1]-vs[0], vs[2]-vs[0])
	d_half = n * (la.norm(n)*th/2)

	vsp = vs+d_half
	vsm = vs-d_half
	vs = np.vstack([vsp,vsm])

	fs = [
		[0,1,2],[5,4,3], # caps
		[0,3,4],[0,4,1], # 0-1 edge
		[1,4,5],[1,5,2], # 1-2 edge
		[2,5,3],[2,3,0]	 # 2-0 edge
	]
	
	return Mesh(vs,fs)

def as_aabb(mesh):
	return aabb_mesh(mesh.vs.min(axis=0), mesh.vs.max(axis=0)+1e-6)

