#!/bin/env python
# coding=utf-8
from __future__ import print_function, division
import sys
import msgpack
import math
import random
import numpy as np
import scipy.linalg as la
import scipy.misc
import scipy.ndimage
import cairo
from tools3d.affine import *
from tools3d.mesh import *
import scene
from scene import Scene, SceneAccel
from index import *
import generate
import localize
import index
import os.path
import traceback
import multiprocessing

import matplotlib
#matplotlib.use('Agg')
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.font_manager as fm


class ArtificialScene(Scene):
	"""
	Artificial cuboid room for experiment.
	X: [-4,4] variable
	Y: [-2,2]
	Z: [-1,1]

	gen_region returns special forms.

	1. almost single point
	2. XY slice
	"""
	def __init__(self, mode='point', half_x=4):
		self.mode = mode

		# create cuboid scene
		vmax = np.array([half_x,2,1], float)
		self.aabb = (-vmax, vmax)
		mesh = scene.aabb_mesh(*self.aabb)

		room_obj = {
			'name': 'room',
			'mesh': mesh.pack_headerless()
		}

		original = {
			'objects': [room_obj]
		}

		Scene.__init__(self, original)		  

	def gen_region(self, unit=0.05):
		if self.mode == 'point':
			return (np.zeros([3],int), np.array([[[True]]]))
		elif self.mode == 'plane':
			vmin, vmax = self.aabb
			h_nx = int(vmax[0]/unit)
			h_ny = int(vmax[1]/unit)

			voxels = np.ones([2*h_nx,2*h_ny,1], bool)
			offset = np.array([-h_nx,-h_ny,0], int)
			return (offset, voxels)
		else:
			raise NotImplementedError('unknown region mode %s specified'%self.mode)

	def get_accel(self):
		return SceneAccel.from_scene_objects(self.objects)

def multi_sum(array, axes):
	red = array
	for i in reversed(list(axes)):
		red = red.sum(axis=i)
	return red

def run_point(quick=True):
	path = [SE3.identity()]
	sample_l = 0.005
	sample_a = math.radians(1)
	lunit = 0.02

	if quick:
		sample_a = math.radians(5)
		aunit = math.radians(30)
		half_xs = [1,2,3,4] # 2,2 is 4-symmetrical case. see mode shift
	else:
		aunit = math.radians(5)
		half_xs = np.linspace(1.5,2.5,num=9)

	path_figure = 'res_point-%d.png'%len(half_xs)

	params = []
	for half_x in half_xs:
		print('== Configuration')
		print(half_x)

		print('== Index')
		scene = ArtificialScene('point',half_x=half_x)

		path_index = generate.index(scene, sample_l=sample_l, sample_a=sample_a,lunit=lunit, aunit=aunit)
		print('index path:', path_index)

		print('== Preprocess')
		direction = index.DiscreteDirection(aunit)
		orientation = index.DiscreteOrientation(aunit)

		grav_const = localize.create_gravity_constraint(np.array([0,0,-1]), direction, orientation)
		down_m = np.array([0,0,-1])
		down_m_i = direction.discretize(down_m)
		ori_filter = grav_const[down_m_i]

		dev = localize.ODeviceModel(index.ErrorModel())
		ind = index.Index.load_for(path_index, dev, ori_filter)

		print('== Evaluation')
		_, [xd], _ = localize.simulate_localize_session(ind, path, dev, base_p='simple-point-hx%f'%half_x, fixed_ori_set=ori_filter)

		for i in range(6):
			red = multi_sum(ind.to_legacy(xd), {0,1,2,3,4,5} - {i})
			print(red)
			if i==4:
				params.append(red)

	print('== Visualization')
	cm_to_inch = 0.393700787
	fig = plt.figure()
	fig.set_size_inches(12*cm_to_inch, 12*cm_to_inch)
	fig.suptitle('Ls=%.3fm As=%.4frad / Lg=%.3fm Ag=%.4frad\n%s'%(
		sample_l, sample_a,
		lunit, aunit,
		dev.get_desc(),
		))
	ax = fig.add_subplot(1,1,1)
	ax.set_xlabel('phi [rad]')
	ax.set_ylabel('W/2 [m]')
	ax.set_xlim(0, 2*math.pi) # phi
	ax.set_ylim(min(half_xs)-0.5,max(half_xs)+0.5) # half X (room dimension)

	for (half_x,dist) in zip(half_xs,params):
		n = len(dist)
		for (i,v) in enumerate(dist):
			if v:
				ax.axhline(half_x, i/n, (i+1)/n, alpha=1, color='g', linewidth=2)

	ax.grid(True)
	fig.savefig(path_figure, dpi=600)
	print('{{%s}}'%path_figure)

def run_plane(quick=True):
	path = [SE3.translate(np.array([1,1,0]))]
	
	if quick:
		sample_a = math.radians(2.5)
		aunit = math.radians(15)
		lunit = 0.2
		sample_l = 0.05
	else:
		sample_a = math.radians(2.5)
		aunit = math.radians(10)
		lunit = 0.2
		sample_l = 0.05

	path_figure = 'res_plane.png'

	print('== Index')
	scene = ArtificialScene('plane',half_x=4)

	path_index = generate.index(scene, sample_l=sample_l, sample_a=sample_a,lunit=lunit, aunit=aunit)
	print('index path:', path_index)

	print('== Preprocess')
	direction = index.DiscreteDirection(aunit)
	orientation = index.DiscreteOrientation(aunit)

	grav_const = localize.create_gravity_constraint(np.array([0,0,-1]), direction, orientation)
	down_m = np.array([0,0,-1])
	down_m_i = direction.discretize(down_m)
	ori_filter = grav_const[down_m_i]

	dev = localize.ODeviceModel(index.ErrorModel())
	ind = index.Index.load_for(path_index, dev, ori_filter)

	print('== Evaluation')
	_, [xd], _ = localize.simulate_localize_session(ind, path, dev, base_p='simple-plane', fixed_ori_set=ori_filter)
	xd = ind.to_legacy(xd)

	for i in range(6):
		red = multi_sum(xd, {0,1,2,3,4,5} - {i})
		print(red)

	xy_dist = multi_sum(xd, {2,3,4,5})
	print(xy_dist)

	phis = np.linspace(0, 2*math.pi, num=orientation.nphi, endpoint=False)
	upper = multi_sum(multi_sum(xd, {2,3,5})[:,xy_dist.shape[1]//2:], {0,1})
	lower = multi_sum(multi_sum(xd, {2,3,5})[:,:xy_dist.shape[1]//2], {0,1})
	print('upper phi', upper)
	print('lower phi', lower)

	print('== Visualization')
	cm_to_inch = 0.393700787
	plt.figure(figsize=(16*cm_to_inch, 24*cm_to_inch))
	plt.suptitle('Ls=%.3fm As=%.4frad / Lg=%.3fm Ag=%.4frad\n%s'%(
		sample_l, sample_a,
		lunit, aunit,
		dev.get_desc(),
		))

	# x-y
	ax_xy = plt.subplot(2,1,1)
	plt.xlabel('x [m]')
	plt.ylabel('y [m]')
	plt.xlim(-4, 4)
	plt.ylim(-2, 2)
	plt.imshow(xy_dist.T[::-1],extent=(-4,4,-2,2),interpolation='nearest')
	plt.grid(True)
	plt.colorbar(ax=ax_xy)
	plt.scatter([1],[1],s=15,facecolor='lightgreen')
	plt.scatter([-1],[-1],s=15,facecolor='lightblue')

	# phi
	plt.subplot(2,1,2)
	plt.xlabel('phi [rad]')
	plt.ylabel('count')
	plt.xlim(0,2*math.pi)
	plt.bar(phis, upper, width=2*math.pi/orientation.nphi, color='lightgreen', label='y>0')
	plt.bar(phis, lower, width=2*math.pi/orientation.nphi, color='lightblue', label='y<0')
	plt.legend()
	plt.grid()

	plt.savefig(path_figure, dpi=600)
	print('{{%s}}'%path_figure)

def vis_slice_gen():
	"""
	See
	https://docs.google.com/spreadsheet/ccc?key=0AqzCoyeuDn-pdHNwYkNZaF85OGNUNmVyRE5YUE5vckE
	for more detail.
	"""
	path_figure = 'plane-slices.png'

	table_small = [
		[3	 ,3.19580549],
		[2.04748951	 ,3.26212222],
		[1.43592424	 ,2.07624169],
		[1.16608852	 ,1.46699292],
		[1.04056521	 ,1.19642168],
		[1.00005005	 ,1.07095741],
		[1	 ,1.0652685],
		[1.03481526	 ,1.18278355],
		[1.15248786	 ,1.43821452],
		[1.40719167	 ,2.00656464],
		[1.97720181	 ,3.73290347],
		[3.74723519	 ,5.18512685],
		[4.03996627	 ,5.25526035],
		[4.03996627	 ,5.81377455],
		[4.03996627	 ,5.90781451],
		[3.49826556	 ,4.37946859],
		[3.12169562	 ,3.58926503],
		[3.00015016	 ,3.21287222],
		[3	 ,3.19580549],
		[3.10444577	 ,3.54835066],
		[3.45746359	 ,4.29355536],
		[3.49826556	 ,4.33586943],
		[3.12169562	 ,3.58926503],
		[3.00015016	 ,3.21287222]]
	table_large = [
		[3.00015016	 ,3.08880795],
		[3.04407564	 ,3.21256407],
		[2.04759199	 ,3.04642446],
		[1.58975008	 ,2.04469944],
		[1.31910608	 ,1.5961483],
		[1.16614688	 ,1.32934957],
		[1.0686208	 ,1.17824564],
		[1.0183954	 ,1.08210978],
		[1.00010011	 ,1.03314725],
		[1.00005005	 ,1.02960265],
		[1.01469188	 ,1.07447284],
		[1.06083917	 ,1.16481471],
		[1.15254555	 ,1.30755892],
		[1.29704276	 ,1.55869937],
		[1.55169403	 ,1.97608092],
		[1.97730078	 ,2.88396251],
		[2.91504911	 ,5.17542246],
		[5.00050054	 ,5.16573623],
		[5.00025026	 ,5.14801325],
		[5.07345939	 ,5.3723642],
		[5.30419585	 ,5.81377455],
		[4.76925023	 ,5.90781451],
		[3.95731825	 ,4.7884449],
		[3.49844065	 ,3.98804872],
		[3.20586239	 ,3.53473692],
		[3.05518619	 ,3.24632934],
		[3.00030032	 ,3.09944174],
		[3.00015016	 ,3.08880795],
		[3.04407564	 ,3.22341852],
		[3.18251751	 ,3.49444413],
		[3.45763664	 ,3.92267677],
		[3.89112829	 ,4.29097142],
		[3.49844065	 ,3.98804872],
		[3.20586239	 ,3.53473692],
		[3.05518619	 ,3.24632934],
		[3.00030032	 ,3.09944174]]

	# 
	table_small = np.array(table_small)
	table_large = np.array(table_large)

	phis_small = np.linspace(0, 2*math.pi, num=len(table_small), endpoint=False)
	phis_large = np.linspace(0, 2*math.pi, num=len(table_large), endpoint=False)

	cm_to_inch = 0.393700787
	plt.figure(figsize=(16*cm_to_inch, 16*cm_to_inch))

	# phi-r
	ax = plt.subplot(1,1,1)
	plt.xlabel(u'φ [rad]')
	plt.ylabel('distance [m]')
	plt.xlim(0,2*math.pi)
	x_labels = ['$0$', '$\\frac{\\pi}{2}$', '$\\pi$', '$\\frac{3}{2}\\pi$', '$2\\pi$']
	ax.set_xticks(np.arange(0,2.5,0.5)*np.pi)
	ax.set_xticklabels(x_labels)

	for (i,(v0,v1)) in enumerate(table_small):
		n = len(table_small)
		plt.axhspan(v0, v1, i/n, (i+1)/n, color='red', alpha=0.5, **({} if i>0 else {'label':'15 deg'}))

	for (i,(v0,v1)) in enumerate(table_large):
		n = len(table_large)
		plt.axhspan(v0, v1, i/n, (i+1)/n, color='blue', alpha=0.5, **({} if i>0 else {'label':'10 deg'}))

	plt.legend()

	plt.grid()

	plt.savefig(path_figure, dpi=600)
	print('{{%s}}'%path_figure)

def run_plane_real(quick=True):
	path_figure = 'res_plane_real.png'

	if quick:
		sample_a = math.radians(2.5)
		aunit = math.radians(15)
		lunit = 0.2
		sample_l = 0.05
	else:
		sample_a = math.radians(2.5)
		aunit = math.radians(10)
		lunit = 0.2
		sample_l = 0.05

	print('== Preprocess')
	direction = index.DiscreteDirection(aunit)
	orientation = index.DiscreteOrientation(aunit)

	grav_const = localize.create_gravity_constraint(np.array([0,0,-1]), direction, orientation)
	down_m = np.array([0,0,-1])
	down_m_i = direction.discretize(down_m)
	ori_filter = grav_const[down_m_i]

	dev = localize.ODeviceModel(index.ErrorModel())
	ind = index.Index.load_for(path_index, dev, ori_filter)

	dev = localize.ODeviceModel(index.ErrorModel())
	ind = index.Index.load_for(path_index, dev, ori_filter)

def vis_depth():
	bases = ['HS-narrow.png', 'HS-wide.png', 'HS-omni.png']
	descs = [u'標準', u'広角', u'全方位']
	margin_h = 5
	yoffset = [50,450,750]
	surf = cairo.ImageSurface(cairo.FORMAT_RGB24, 400*4+margin_h*3, 1000)
	ctx = cairo.Context(surf)
	ctx.set_source_rgb(1,1,1)
	ctx.paint()
	for (i,base) in enumerate(bases):
		ctx.save()
		ctx.translate(0,yoffset[i])

		ctx.save()
		ctx.scale(5,5)
		ctx.set_source_rgb(0,0,0)
		ctx.select_font_face('VL Gothic')
		ctx.save()
		ctx.translate(0,-1.5)
		ctx.text_path(descs[i])
		ctx.fill()
		ctx.restore()

		ctx.restore()

		for j in range(4):
			ctx.save()
			ctx.translate((400+margin_h)*j, 0)

			path = '%spose%03d-render.png'%(base, j*25)
			print(path)
			s = cairo.ImageSurface.create_from_png(path)

			if i<2:
				ctx.scale(-400/s.get_width(), -400/s.get_width())
				ctx.translate(-s.get_width(),-s.get_height())
				ctx.set_source_surface(s)
				ctx.paint()
			else:
				ctx.scale(400/s.get_width(), -400/s.get_width())
				ctx.translate(0,-s.get_height())
				ctx.set_source_surface(s)
				ctx.paint()

			ctx.restore()
		ctx.restore()

	for i in range(4):
		ctx.set_source_rgb(0,0,0)
		ctx.save()
		ctx.translate(400*i, 1000)
		ctx.scale(5,5)
		ctx.select_font_face('VL Gothic')
		ctx.text_path('t=%.1fs'%(i*0.1*25))
		ctx.fill()
		ctx.restore()

	surf.write_to_png('composite.png')


def eval_single(param):
	try:
		scene = Scene.load('conference.annot.mp')

		param = param.copy()
		print('processing',param)

		ix_path = generate.index(scene,
			sample_l=param['Ls'], sample_a=param['As'],
			lunit=param['Lg'], aunit=param['Ag'])

		p_fig = 'Gs-%.3fm-%.3frad.png'%(param['Lg'],param['Ag'])
		lval = localize.visualize_localization(ix_path, p_fig, 'omni', aunit=param['Ag'])
		print('==*= %s : %f'%(param,lval))

		return {
			'Ls': param['Ls'],
			'As': param['As'],
			'Lg': param['Lg'],
			'Ag': param['Ag'],
			'lval': lval
		}

	except:
		traceback.print_exc()
		return {
			'Ls': param['Ls'],
			'As': param['As'],
			'Lg': param['Lg'],
			'Ag': param['Ag'],
			'exception': traceback.format_exc()
		}

def run_grid_change():
	try:
		ls = list(msgpack.unpack(open('grid-results.mp','rb')))
	except IOError:
		ls = []

	print('starting from',len(ls),'points')

	# generate params
	params = []
	for lg in [0.3,0.4,0.5,0.6,0.7,0.8]:
		for ag in [0.3,0.4,0.5,0.6]:
			exists = False
			for l in ls:
				if l['Lg']==lg and l['Ag']==ag and 'exception' not in l:
					exists = True
					break

			if exists:
				print('already exists. skipping')
			else:
				params.append({
					'Ls': 0.05,
					'As': 0.12, #0.07,
					'Lg': lg,
					'Ag': ag,
				})
	print(params)

	ls += multiprocessing.Pool(processes=3).map(eval_single, params)
	msgpack.pack(ls, open('grid-results.mp','wb'))
	print(len(ls),'points written')

def vis_grid_change():
	path_figure = 'grid.png'

	ls = msgpack.unpack(open('grid-results.mp','rb'))

	cm_to_inch = 0.393700787
	plt.figure(figsize=(16*cm_to_inch, 16*cm_to_inch))

	delta_l={
		0.3: 0.81,
		0.4: 0.91,
		0.5: 0.95,
		0.6: 0.94,
		0.7: 0.83,
		0.8: 0.29
	}
	# generate voronoii
	cmap = cm.get_cmap('Blues')
	#cmap.vlim(12,22)
	img = np.zeros([200,100])
	pts = [((l['Lg'],l['Ag']),l['lval']+delta_l[l['Lg']]) for l in ls if 'exception' not in l and not math.isnan(l['lval'])]
	for iy in range(200): # 0-1
		for ix in range(100): # 0-1
			p = np.array([ix*0.01,iy*0.0025+0.1]) # L,A
			img[iy,ix] = min(pts,key=lambda q:la.norm(p-q[0]))[1]
	#sc = 

	img = scipy.ndimage.gaussian_filter(img, 5)
			
	#plt.imshow(img[::-1], extent=(0,1,0.1,0.35),cmap=cmap)



	xs = []
	ys = []
	vs = []
#	 cmap = cm.get_cmap()
	for l in ls:
		if 'exception' not in l:
			xs.append(l['Lg'])
			ys.append(l['Ag'])
			vs.append(l['lval']+delta_l[l['Lg']])
			print(l)


	# #plt.xscale('log')
	# #plt.yscale('log')
	# 
	plt.xlim(0.3,0.8)
	plt.ylim(0.15,0.3)
	plt.xlabel('Lg [m]')
	plt.ylabel('Ag [rad]')
	plt.grid()

	#cmap.set_clim(10,22)
	sc = plt.scatter(xs,ys,c=vs,cmap=cmap,edgecolor=None,s=100)
	sc.set_clim(10,23)

	cs_c=plt.contour(np.linspace(0,1,num=100),np.linspace(0.1,0.6,num=200),img, range(10,23),cmap=cmap)
	cs_c.set_clim(10,23)

#	 locs = [(0,0)] * 14
	plt.clabel(cs_c,fmt='%.0f',fontsize=10,manual=True)

	cb=plt.colorbar(sc)
	cb.set_label('$\mathcal{L} \mathrm{[bit]}$')


	for i in range(7):
		lgs = np.linspace(0.3,0.8)

		ags = (4*math.pi**3 * 80 / 2**(20+i))**(1/3) / lgs
		cs_p = plt.plot(lgs,ags,ls='--',label='Lmax=%d'%(20+i)) #,c=cmap.get_rgba(20+i))
		#cs_p.set_clim(10,23)

		#ags = (2*math.pi**2 / 2**(5+i))**0.5 + lgs*0
		#plt.plot(lgs,ags)
	plt.legend(loc=2,prop={'size':8}).get_frame().set_alpha(0.8)

	#

	# flush
	plt.savefig(path_figure, dpi=600)
	print('{{%s}}'%path_figure)

if __name__ == '__main__':
	run_point()
