"""
Spherical geometry
"""
from __future__ import print_function, division
import math
import numpy as np
import scipy.linalg as la
from spacetime import parallel
from tools3d.affine import SE3

class Arc(object):
	"""
	An arc is a subset of a great circle.
	Arc also has a side.
	"""
	def __init__(self, n, p0, angle):
		assert(np.isfinite(n).all())

		self.p0 = p0
		if angle>=0:
			self.n = n
			self.angle = angle
		else:
			self.n = -n
			self.angle = - angle

	@staticmethod
	def from_nearest(p,q):
		"""
		Nearest great arc between p and q.
		Left side of path p->q will be True side.
		"""
		n = np.cross(p,q)
		if la.norm(n)<1e-6:
			raise ValueError('cannot create arc from %s and %s'%(p,q))
		
		n /= la.norm(n)
		return Arc(n, p, math.acos(max(-1,min(1,np.dot(p,q)))))

	def intersect_p(self, other):
		"""
		Return intersection point or None.

		Two different great circles always intersect at 2 points.
		Check if one of the points are included in both arcs.
		"""
		pole = np.cross(self.n, other.n)
		pole /= la.norm(pole)
		for p in [pole,-pole]:
			if self.on_arc(p) and other.on_arc(p):
				return p
		return None

	def to_circle(self):
		return Arc(self.n, self.p0, 2*math.pi)
	
	def on_arc(self, p):
		""" Check if p on the same great circle is on the arc. """
		return self.to_angle(p)<=self.angle

	def side(self, p):
		return np.dot(p,self.n)>=0

	def to_angle(self, p):
		"""
		Return angle in [0,2pi) for p on the same great circle.
		(if p is not on the great circle, result is undefined)
		"""
		cos = np.dot(self.p0, p)
		sin = np.dot(self.n, np.cross(self.p0, p))
		return (math.atan2(sin, cos)+2*math.pi)%(2*math.pi)

	def solidify(self):
		m = int(max(1,self.angle/0.02))
		dR = SE3.rotate(self.n, self.angle/m)

		ps = []
		p = self.p0
		for i in range(m):
			ps.append(p)
			p = dR.apply(p)

		return parallel(ps)

	def __repr__(self):
		return str([self.n,self.p0,self.angle])


class Triangle(object):
	"""
	A triangle on a sphere.
	It is composed of 3 great arcs.

	inside of triangle is True.
	"""
	def __init__(self, *ps):
		""" ps must be in CCW order. """
		assert(len(ps)==3)
		self.ps = ps
		try:
			self.arcs = self.calc_arcs()
		except ValueError:
			raise ValueError('invalid spherical triangle %s'%str(self.ps))

	def intersect(self, other):
		""" 
		Check if 2 triangles intersects.
		"""
		if any(a0.intersect_p(a1)!=None for a0 in self.arcs for a1 in other.arcs):
			return True
		if other.contains(self.ps[0]):
			return True
		if self.contains(other.ps[0]):
			return True
		return False

	def get_arcs(self):
		return self.arcs

	def contains(self, p):
		return all(arc.side(p) for arc in self.arcs)

	def calc_arcs(self):
		return [Arc.from_nearest(self.ps[i],self.ps[(i+1)%3]) for i in range(3)]

	def solidify(self):
		return parallel([arc.solidify() for arc in self.arcs])

	def __repr__(self):
		return str([self.arcs])