"""
Two main process of voxelize.py is:
* mesh -> voxel(tri-value)
* voxel(tri-value) -> topological voxel

This test covers the latter.
"""
from __future__ import print_function, division
import unittest
import voxelize as vx
import numpy as np

class FillTestCase(unittest.TestCase):
	def test_fill_zero(self):
		arr = np.ones([5,6,7], bool)
		n_labels, larr = vx.fill_area(arr)
		self.assertEqual(n_labels, 0)
		self.assertTrue((larr == 255).all()) # all voxel is wall
	
	def test_fill_single(self):
		arr = np.zeros([5,6,7], bool)
		n_labels, larr = vx.fill_area(arr)
		self.assertEqual(n_labels, 1)
		self.assertTrue((larr == 0).all()) # all voxel is label 0

	def test_fill_two(self):
		arr = np.zeros([3,3,3], bool)
		arr[:,1,:] = True # put wall at Y=1

		n_labels, larr = vx.fill_area(arr)
		self.assertEqual(n_labels, 2)
		self.assertTrue((larr[:,0,:]==0).all()) # region 0
		self.assertTrue((larr[:,1,:]==255).all()) # wall
		self.assertTrue((larr[:,2,:]==1).all()) # region 1

class TopologicalTestCase(unittest.TestCase):
	def test_empty(self):
		""" Check returned shape & emptiness for a single empty cell. """
		arr = np.empty([2,2,2], np.uint8)
		arr[:] = vx.S_EMPTY

		tv = vx.downsample(arr, 2)
		self.assertEqual(len(tv), 1)
		self.assertEqual(len(tv[0]), 1)
		self.assertEqual(len(tv[0][0]), 1)

		cell = tv[0][0][0]
		self.assertEqual(len(cell), 1) # single region

		region = cell[0]
		self.assertTrue(region['empty'])
		self.assertEqual(region['connections'], [[]]*6)

	def test_half(self):
		""" Check returned non-emptiness of a single half cell. """
		arr = np.empty([2,2,2], np.uint8)
		arr[:] = vx.S_HALF

		cell = vx.downsample(arr, 2)[0][0][0]
		self.assertEqual(len(cell), 1) # single region

		region = cell[0]
		self.assertFalse(region['empty']) # NOT empty
		self.assertEqual(region['connections'], [[]]*6)

	def test_empty2(self):
		""" Check connection of two empty cells, ----. """
		arr = np.empty([2*2,2,2], np.uint8)
		arr[:] = vx.S_EMPTY

		tv = vx.downsample(arr, 2)
		self.assertEqual(len(tv), 2) # two cells along X-axis

		region0 = tv[0][0][0][0]
		region1 = tv[1][0][0][0]

		self.assertTrue(region0['empty'])
		self.assertTrue(region1['empty'])
		self.assertEqual(region0['connections'], [[],[0], [],[], [],[]]) # 0:0 --(X+)-> 1:0
		self.assertEqual(region1['connections'], [[0],[], [],[], [],[]]) # 0:0 <-(X-)-- 1:0
	
	def test_separate2(self):
		""" Check (absence of connection) of two cells, -|-- (-:empty, |:wall). """
		arr = np.empty([2*2,2,2], np.uint8)
		arr[:] = vx.S_EMPTY
		arr[1,:,:] = vx.S_FILLED # put wall in cell0

		tv = vx.downsample(arr, 2)
		region0 = tv[0][0][0][0]
		region1 = tv[1][0][0][0]

		self.assertFalse(region0['empty']) # cell0 contains wall
		self.assertTrue(region1['empty']) # cell1 is empty
		self.assertEqual(region0['connections'], [[]]*6)
		self.assertEqual(region1['connections'], [[]]*6)

	def test_pane(self):
		"""
		Check two regions created by wall running through two cells.

		cell0 has region 0 and 1. cell1 also has region 0 and 1.
		Connections should be (0,0)<->(1,0) and (0,1)<->(1,1), but no others.
		"""
		arr = np.empty([2*3,3,3], np.uint8)
		arr[:] = vx.S_EMPTY
		arr[:,1,:] = vx.S_FILLED # put wall at Y=1. This will divide both cells into 2 regions.

		tv = vx.downsample(arr, 3)
		cell0 = tv[0][0][0]
		cell1 = tv[1][0][0]
		self.assertEqual(len(cell0), 2)
		self.assertEqual(len(cell1), 2)

		for regid in [0,1]: # overly strong. (flipping of region ids in cell1 is ok), but current impl. is deterministic, so it's ok.
			self.assertEqual(cell0[regid]['connections'], [[],[regid], [],[], [],[]]) # 0:i --(X+)-> 1:i
			self.assertEqual(cell1[regid]['connections'], [[regid],[], [],[], [],[]]) # 0:i <-(X-)-- 1:i

	def test_pane1(self):
		"""
		Check one region-two regions connections created by wall in cell1.

		cell0 is empty. cell1 has region 0 and 1.
		Connections should be (0,0)<->(1,0) and (0,0)<->(1,1).
		"""
		arr = np.empty([2*3,3,3], np.uint8)
		arr[:] = vx.S_EMPTY
		arr[3:,1,:] = vx.S_FILLED # put wall at Y=1. This will divide cell1 into 2 regions.

		tv = vx.downsample(arr, 3)
		cell0 = tv[0][0][0]
		cell1 = tv[1][0][0]
		self.assertEqual(len(cell0), 1)
		self.assertTrue(cell0[0]['empty'])
		self.assertEqual(len(cell1), 2)

		self.assertEqual(cell0[0]['connections'], [[],[0,1], [],[], [],[]]) # 0:0 --(X+)-> 1:i
		self.assertEqual(cell1[0]['connections'], [[0],[], [],[], [],[]]) # 0:0 <-(X-)-- 1:0
		self.assertEqual(cell1[1]['connections'], [[0],[], [],[], [],[]]) # 0:0 <-(X-)-- 1:1
