#!/bin/env python
from __future__ import print_function, division
import numpy as np
import scipy.linalg as la
import scipy.ndimage as ndimage
import math
import sys
import numpy as np
import time
import types
import random
from tools3d.mesh import Mesh
from tools3d.affine import *
from spacetime import parallel
import spherical
import json
from scene import Scene
from Queue import Queue
from fvxl import dumpSimpleFVXL, loadSimpleFVXL

S_EMPTY = 0
S_HALF = 0x80
S_FILLED = 0xff

def voxelize_surface(scene, unit):
	""" Returned voxels contain either S_EMPTY or S_HALF. """
	# pass 1: analyze scene
	ttris_scene = [(obj['aabb'],obj['mesh'].flatten()) for obj in scene.objects]
	vmin = np.array([1e3,1e3,1e3])
	vmax = -vmin
	for (o_aabb, _) in ttris_scene:
		vmin = np.minimum(vmin, o_aabb[0])
		vmax = np.maximum(vmax, o_aabb[0])
	size = vmax-vmin
	shape = tuple(np.floor(size/unit).astype(int)+1)
	print('AABB',vmin,vmax)
	print('voxel',shape)

	# pass 2: write
	vscene = np.empty(shape, np.uint8)
	vscene[:] = S_EMPTY
	for (i_obj,(o_aabb, o_tris)) in enumerate(ttris_scene):
		print("processing object %d/%d"%(i_obj+1, len(ttris_scene)))
		for tri in o_tris:
			org = tri[0]
			e1 = tri[1]-org
			e2 = tri[2]-org
			n1 = int(2*la.norm(e1)/unit)+1
			n2 = int(2*la.norm(e2)/unit)+1
			for i1 in range(n1+1):
				for i2 in range(n2+1):
					t1 = i1/n1
					t2 = i2/n2
					if t1+t2>1:
						continue

					p = org + e1*t1 + e2*t2
					try:
						vscene[tuple(map(int,np.floor((p-vmin)/unit)))] = S_HALF
					except IndexError:
						pass

	return vscene

def fix_volume(volume):
	""" Convert surface volume to sane volume using heuristics """
	surface = volume==S_HALF
	filled = ndimage.binary_fill_holes(surface)

	volume[:] = S_FILLED
	volume = np.where(np.logical_not(filled), S_EMPTY, volume)
	volume = np.where(surface, S_HALF, volume)
	return volume

def _fill_binary(arr, i, output=None):
	""" output can be arr """
	assert(not arr[i])
	if output==None:
		output = arr.copy()

	sx,sy,sz = arr.shape

	q = Queue()
	q.put(i)
	while not q.empty():
		i = q.get()
		x,y,z = i

		if x>=0 and y>=0 and z>=0 and x<sx and y<sy and z<sz and not output[i]:
			output[i] = True
			q.put(i)

			q.put((x+1,y,z))
			q.put((x-1,y,z))
			q.put((x,y+1,z))
			q.put((x,y-1,z))
			q.put((x,y,z+1))
			q.put((x,y,z-1))

	return output

def fill_area(arr):
	"""
	Convert boolean (true:filled, false:empty) voxel to labeled voxels
	based on 6-connectivity.

	Return (number of labels, labeled voxel)
	labels will be 0,1,...
	wall will be 255
	"""
	lvoxel = arr.astype(np.uint8)*255 # set walls
	filled = lvoxel.copy()

	lid = 0
	while not filled.all():
		before = filled.copy()
		i = np.unravel_index(filled.argmin(), filled.shape) # find smallest non-filled index
		_fill_binary(filled, i, filled)

		newly_filled_volume = filled != before
		lvoxel = np.where(newly_filled_volume, lid, lvoxel)
		lid += 1

	return (lid,lvoxel)
		


def downsample(arr, n):
	"""
	Get coarser "topological voxel" representation from
	finer tri-valued voxels by n^3 downsampling.

	Internal:
	common order of direction scan is X-,X+,Y-,Y+,Z-,Z+.
	"""
	def extract(ix,iy,iz):
		return arr[ix*n:(ix+1)*n, iy*n:(iy+1)*n, iz*n:(iz+1)*n]

	def faces(lvoxel):
		return [
			lvoxel[0], lvoxel[-1],
			lvoxel[:,0], lvoxel[:,-1],
			lvoxel[:,:,0], lvoxel[:,:,-1]]

	shape = tuple(norg//n for norg in arr.shape)

	# generate regions
	rvoxels = []
	for ix in xrange(shape[0]):
		print(ix,shape[0])
		row_y = []
		for iy in xrange(shape[1]):
			row_z = []
			for iz in xrange(shape[2]):
				voxel = extract(ix,iy,iz)
				if (voxel==S_EMPTY).all():
					rvoxel = (1,True,faces(voxel*0))
				elif (voxel==S_FILLED).all():
					rvoxel = (0,False,faces(voxel*0+255))
				elif (voxel!=S_FILLED).all():
					rvoxel = (1,False,faces(voxel*0))
				else:
					cnt, lvoxel = fill_area(voxel==S_FILLED)
					assert(0<cnt)
					if cnt>=255:
						raise NotImplementedError('unsupportedly many regions found')
					rvoxel = (cnt,False,faces(lvoxel))

				row_z.append(rvoxel)
			row_y.append(row_z)
		rvoxels.append(row_y)

	# calculate connections
	def valid_ix(i):
		return all(0<=k for k in i) and all(k<s for (k,s) in zip(i,shape))

	def get_adjacent_faces(i):
		"""
		Get faces of adjacent voxels.
		Return dummy wall face when there's no neighbor.
		"""
		ix,iy,iz = i

		ins = [
			(ix-1,iy,iz),(ix+1,iy,iz),
			(ix,iy-1,iz),(ix,iy+1,iz),
			(ix,iy,iz-1),(ix,iy,iz+1)]

		results = []
		for (ix,j) in enumerate(ins):
			fix = ix ^ 1 # invert direction
			if valid_ix(j):
				nls,_,faces = rvoxels[j[0]][j[1]][j[2]]
				results.append((nls,faces[fix]))
			else:
				results.append((1,np.ones([n,n],np.uint8)*255))

		return results
		

	tvoxels = []
	for ix in xrange(shape[0]):
		print(ix,shape[0])
		row_y = []
		for iy in xrange(shape[1]):
			row_z = []
			for iz in xrange(shape[2]):
				n_labels,is_empty,faces = rvoxels[ix][iy][iz]
				faces_n = get_adjacent_faces((ix,iy,iz))

				if is_empty:
					tvoxel = [
						{
							'empty': True,
							'connections': [
								[j for j in range(nls) if (fn==j).any()]
									for (nls,fn) in faces_n]
						}
					]
				else:
					tvoxel = []
					for i in range(n_labels):
						tvoxel.append({
							'empty': False,
							'connections': [
								[j for j in range(nls) if ((f==i) & (fn==j)).any()]
								for (f,(nls,fn)) in zip(faces, faces_n)]
							})

				row_z.append(tvoxel)
			row_y.append(row_z)
		tvoxels.append(row_y)

	return tvoxels
	


def main():
	"""
	topology = topo . volume
	"""
	if len(sys.argv) == 6 and sys.argv[1] == 'topology':
		path_scene, s_unit, ds_rate, path_topo = sys.argv[2:]

		scene = Scene.load(path_scene)
		voxel = fix_volume(voxelize_surface(scene, float(s_unit)))
		top_voxel = downsample(fix_volume(voxelize_surface(scene)), int(ds_rate))
		json.dump(top_voxel, open(path_topo,'w'))
	elif len(sys.argv) == 5 and sys.argv[1] == 'volume':
		path_scene, s_unit, path_fvxl = sys.argv[2:]

		scene = Scene.load(sys.argv[2])
		voxel = fix_volume(voxelize_surface(scene, float(s_unit)))
		dumpSimpleFVXL(voxel, path_fvxl)
	elif len(sys.argv) == 5 and sys.argv[1] == 'topo':
		path_fvxl, ds_rate, path_topo = sys.argv[2:]

		voxel = loadSimpleFVXL(path_fvxl)
		top_voxel = downsample(voxel, int(ds_rate))
		json.dump(top_voxel, open(path_topo,'w'))
	else:
		print('topology = topo . volume')
		print('usage: ./voxelize.py topology <scene path> <unit in meter> <downsampling rate> <TOPO JSON path>')
		print('usage: ./voxelize.py volume <scene path> <unit in meter> <FVXL path>')
		print('usage: ./voxelize.py topo <FVXL path> <downsampling rate> <TOPO JSON path>')


if __name__ == '__main__':
	main()
